from functools import wraps


def awaitify(sync_func):
    """
    Wrap a synchronous callable to allow ``await``'in it
    Since python 3.8 asyncio.coroutine is deprecated
    TODO: Check if there's a proper way to rewrite all sync
    funcs (that calls Event)into native coroutines
    """
    @wraps(sync_func)
    async def async_func(*args, **kwargs):
        return sync_func(*args, **kwargs)
    return async_func



import argparse

from functools import partial
from pathlib import Path

import requests

from pypeman.tools.view_store import Filter
from pypeman.tools.view_store import process_file_store


def http_send(url, store_id, msg):
    """
    sends message to a given url
    """
    print(f"send {store_id} to {url}")
    print(f"msg = {msg}")
    payload = msg.payload
    ses = requests.session()
    ses.post(url, json=payload)


def get_send_function(dst_str):
    """
    determines function, that will send a message to a target
    """
    if dst_str.startswith("http"):
        return partial(http_send, dst_str)

    raise Exception(f"can't determine send function for {dst_str}")


def mk_parser():
    """
    create argument parser for send_from_store command
    """
    descr = "send store contents to an endpoint"
    parser = argparse.ArgumentParser(descr)
    parser.add_argument("path", help="path to search messages")
    parser.add_argument(
        "--type", "-t",
        default="json",
        help="payload type (dflt = %(default)s)",
    )
    parser.add_argument(
        "--filter", "-f",
        help="filter (dot_separated_names=value)",
        action="append",
    )
    parser.add_argument(
        "--destination", "-d",
        help="destination 'url' to send messages to",
        default="http://localhost:8000",
    )
    return parser


def main():
    options = mk_parser().parse_args()
    print(f"options {options}")
    path = Path(options.path)
    filters = options.filter or []
    filters = [Filter(filter_str) for filter_str in filters]
    send_func = get_send_function(options.destination)
    process_file_store(path, pl_type=options.type, filters=filters, action=send_func)


if __name__ == "__main__":
    main()



from aiohttp import web

from pypeman.plugins.remoteadmin import views


def init_urls(app, prefix=""):
    """
    Create the pypeman remoteadmin routing

    Args:
        app (aiohttp.web.Application): The aiohttp web app where the
                                        url routings have to be added
    """
    app.add_routes([
        # API :
        web.get(prefix + '/channels', views.list_channels),
        web.get(prefix + '/channels/{channelname}/start', views.start_channel),
        web.get(prefix + '/channels/{channelname}/stop', views.stop_channel),
        web.get(prefix + '/channels/{channelname}/messages', views.list_msgs),
        web.get(prefix + '/channels/{channelname}/messages/{message_id}/replay', views.replay_msg),
        web.get(prefix + '/channels/{channelname}/messages/{message_id}/view', views.view_msg),
        web.get(prefix + '/channels/{channelname}/messages/{message_id}/preview', views.preview_msg),
        # WEBSOCKETS :
        web.get(prefix + '/', views.backport_old_client),
    ])



"""
Helpers / fixtures for pytest
"""

import pytest

import pypeman.channels
import pypeman.endpoints
import pypeman.nodes


@pytest.fixture(scope="function")
def clear_graph():
    """
    ensure, that before and after a pypeman test
    the pypeman graph is entirely cleared
    """
    n_nodes = len(pypeman.nodes.all_nodes)
    n_channels = len(pypeman.channels.all_channels)
    n_endpoints = len(pypeman.endpoints.all_endpoints)
    pypeman.nodes.reset_pypeman_nodes()
    pypeman.channels.reset_pypeman_channels()
    pypeman.endpoints.reset_pypeman_endpoints()
    yield n_nodes, n_channels, n_endpoints
    pypeman.nodes.reset_pypeman_nodes()
    pypeman.channels.reset_pypeman_channels()
    pypeman.endpoints.reset_pypeman_endpoints()



import json
import logging
import ssl
import re

import aiohttp

from aiohttp import web

from pypeman import endpoints, channels, nodes, message
from pypeman.errors import PypemanParamError


logger = logging.getLogger(__name__)

# Regex to extract dynamic params from a string to permits complex search
# in nested dicts:
#
# Example:
# url = "toto/tutu/%(gigi.gogo)s/bla/%(rigo)r/toto"
# Yields 2 results: "gigi.gogo" and "rigo"
str_named_param_regex = re.compile(r"%\((?P<keyval>[^\)]*)\)[r|s|d|]")

# Regex used to split a string by not escaped "."
# example:
# "titi.toto.tutu" yields 3 results: "titi", "toto", "tutu"
# "titi.toto\.tutu" yields 2 results: "titi", "toto\.tutu" (the \ will be removed in code)
not_escaped_dot_regex = re.compile(r"(?<!\\)\.")


class HTTPEndpoint(endpoints.SocketEndpoint):
    """
    Endpoint to receive HTTP connection from outside.

    """
    def __init__(
            self,
            adress=None, address=None, port=None,  # obsolete params
            loop=None,
            http_args=None,
            host=None,
            sock=None,  # TODO: Why have sock and port if it's the same ?
            reuse_port=None,
            ):
        """
            :param http_args: dict to pass as **kwargs** to aiohttp.Application for example for
                `client_max_size`
            :param reuse_port: bool if true then the listening port specified in the url parameter)
                will be shared with other processes on same port
            :param host: string 'host:port' or ':port' or 'host'
            :param sock: host-string (same as host parameter)
                or socket-string ("unix:/sojet/file/path")
                or bound socket object
        """

        self.http_args = http_args or {}
        self.ssl_context = self.http_args.pop('ssl_context', None)
        self._app = None

        address = address or adress
        if address or port:
            raise PypemanParamError(
                "HTTPEndpoint Obsolete params ('address', 'port') ")

        if host and sock:
            raise PypemanParamError("There can only be one (parameter host or sock)")
        sock = sock or host or ''

        super().__init__(loop=loop, sock=sock, reuse_port=reuse_port)

    def add_route(self, *args, **kwargs):
        """
        Adds a route to the http server.
        This is normally called when an http channel is added to this endpoint
        """
        if self._app is None:
            self._app = web.Application(**self.http_args)

        self._app.router.add_route(*args, **kwargs)

    async def start(self):
        self.make_socket()
        if self._app is not None:
            srv = await self.loop.create_server(
                protocol_factory=self._app.make_handler(),
                sock=self.sock_obj,
                ssl=self.ssl_context,
            )
            message = "Server started at %r" % self.sock
            print(message)
            logger.info(message)
            return srv
        else:
            print("No HTTP route.")
            logger.warning("No HTTP route.")


class HttpChannel(channels.BaseChannel):
    """
    Channel that handles Http requests to a given path of the url.
    The Http message is the message payload and some headers
    become metadata of message. Needs ``aiohttp`` python dependency to work.

    :param endpoint: HTTP endpoint the channel will be using. ( scheme://fqdn:port for example)
    :param method: Method filter.
    :param url: Only matching urls messages will be sent to this channel.
    :param encoding: Encoding of message. Default to 'utf-8'.
    :param add_headers: if True headers will be added to the message meta data

    """
    app = None

    def __init__(
        self,
        *args,
        endpoint=None,
        method='*',
        url='/',
        encoding=None,
        add_headers=False,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.add_headers = add_headers
        self.method = method
        self.url = url
        self.encoding = encoding
        if endpoint is None:
            raise TypeError('Missing "endpoint" argument')
        self.http_endpoint = endpoint

    async def start(self):
        first_start = self._first_start
        await super().start()
        if first_start:
            self.http_endpoint.add_route(self.method, self.url, self.handle_request)

    async def handle_request(self, request):
        content = await request.text()

        # extract match info from iohttp request
        meta = dict(request.match_info)

        # match infos will be overwritten if coinciding with known keywords like
        # 'url', 'method', . . .
        meta.update({
            'method': request.method,
            'url': str(request.url),
            'get_params': dict(request.query),
            })
        if self.add_headers:
            meta["headers"] = [(hdr, val) for (hdr, val) in request.headers.items()]

        msg = message.Message(content_type='http_request', payload=content, meta=meta)
        try:
            result = await self.handle(msg)
            encoding = self.encoding or 'utf-8'
            return web.Response(
                body=str(result.payload).encode(encoding),
                status=result.meta.get('status', 200),
                content_type=getattr(result, "content_type", None),
            )

        except channels.Dropped:
            return web.Response(body="Dropped".encode('utf-8'), status=200)
        except Exception as exc:
            logger.exception(
                "HttpChannel %s raise an error, cannot send 200 speed response, will return 503",
                str(self))
            return web.Response(body=str(exc).encode('utf-8'), status=503)


class HttpRequest(nodes.BaseNode):
    """
        Http request node
        :param url: url to send.
        :param method: 'get', 'put' or 'post', use meta['method'] if None, Default to 'get'.
        :param headers: headers for request, use meta.get('headers') if None.
        :param cookies: cookies for request, use meta.get('cookies') if None.
        :param auth: tuple or aiohttp.BasicAuth object.
        :param verify: verify ssl. Default True.
        :param params: get params in dict. List for multiple elements, ex :
                       {'param1': 'omega', param2: ['alpha', 'beta']}
                       The key can be a function that takes a msg as input param
        :param client_cert: tuple with .crt and .key path
        :param binary: bool, Get response content as bytes
        :param send_as_json: bool, If the method is a PATCH/POST/PUT, send data as json
        :param json: bool, Parse Json response content
        # TODO maybe add an auto parser if for example Content-Type header is application/json
    """

    def __init__(self, url, *args, method=None, headers=None, auth=None,
                 verify=True, params=None, client_cert=None, cookies=None,
                 binary=False, json=False, send_as_json=False, old_url_parsing=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = url
        self.method = method
        self.headers = headers
        self.cookies = cookies
        if isinstance(auth, tuple):
            auth = aiohttp.BasicAuth(*auth)
        self.auth = auth
        self.verify = verify
        self.params = params
        self.client_cert = client_cert
        self.url = self.url.replace('%(meta.', '%(')  # TODO: why ???
        self.payload_in_url_dict = 'payload.' in self.url  # TODO: should I remove ?
        self.params_in_url = str_named_param_regex.findall(self.url)
        self.binary = binary
        self.json = json
        self.send_as_json = send_as_json
        self.old_url_parsing = old_url_parsing
        # TODO: create used payload keys for better perf of generate_request_url()

    def generate_request_url(self, msg):
        request_url = self.url
        if self.old_url_parsing:
            url_dict = msg.meta
            if self.payload_in_url_dict:
                url_dict = dict(url_dict)
                try:
                    for key, val in msg.payload.items():
                        url_dict['payload.' + key] = val
                except AttributeError:
                    self.channel.logger.error(
                        "Payload must be a python dict if used to generate url. "
                        "This can be fixed using JsonToPython node before your "
                        "RequestNode")
                    raise
            try:
                request_url = self.url % url_dict
            except Exception as exc:
                logger.error("cannot create url %r with args %r", self.url, repr(url_dict))
                raise exc
        else:
            # New recursive params parsing
            if self.params_in_url:
                params = {}
                for param in self.params_in_url:
                    subparams = not_escaped_dot_regex.split(param)
                    if subparams[0] == "payload":
                        data = msg.payload
                        subparams.pop(0)
                    elif subparams[0] == "meta":
                        data = msg.meta
                        subparams.pop(0)
                    else:
                        data = msg.meta
                    for subparam in subparams:
                        subparam = subparam.replace(r"\.", ".")
                        data = data[subparam]
                    params[param] = data
                try:
                    request_url = self.url % params
                except Exception as exc:
                    logger.error("cannot create url %r with args %r", self.url, repr(params))
                    raise exc
        return request_url

    async def handle_request(self, msg):
        """ generate url and handle request """
        url = self.generate_request_url(msg)
        loop = self.channel.loop
        conn = None
        ssl_context = None
        if self.client_cert:
            if self.verify:
                ssl_context = ssl.create_default_context()
            else:
                ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                ssl_context.check_hostname = False
                ssl_context.verify_mode = ssl.CERT_NONE
            try:
                ssl_context.load_cert_chain(self.client_cert[0], self.client_cert[1])
            except FileNotFoundError:
                logger.error("loading certs %s failed", self.client_cert)
                raise
            conn = aiohttp.TCPConnector(ssl=ssl_context, loop=loop)
        else:
            conn = aiohttp.TCPConnector(ssl=self.verify, loop=loop)

        headers = nodes.choose_first_not_none(self.headers, msg.meta.get('headers'))
        cookies = nodes.choose_first_not_none(self.cookies, msg.meta.get('cookies'))
        method = nodes.choose_first_not_none(self.method, msg.meta.get('method'), 'get')
        params = nodes.choose_first_not_none(self.params, msg.meta.get('params'))

        get_params = None
        if params:
            params = params.copy()
            get_params = []
            for key, param in params.items():
                if callable(param):
                    param = param(msg)
                if isinstance(param, list):
                    for value in param:
                        get_params.append((key, value))
                else:
                    get_params.append((key, param))

        basic_auth = self.auth

        data = None
        if method.lower() in ['put', 'post', 'patch']:
            data = msg.payload
        async with aiohttp.ClientSession(
            connector=conn,
            cookies=cookies,
            trust_env=True,  # respect e.g. http_proxy env vars
        ) as session:
            if self.send_as_json:
                resp = await session.request(
                        method=method,
                        url=url,
                        auth=basic_auth,
                        headers=headers,
                        params=get_params,
                        json=data
                        )
            else:
                resp = await session.request(
                        method=method,
                        url=url,
                        auth=basic_auth,
                        headers=headers,
                        params=get_params,
                        data=data
                        )
            if self.binary:
                setattr(resp, "content", await resp.read())
            else:
                setattr(resp, "content", str(await resp.text()))
        return resp

    async def process(self, msg):
        """ handles request """
        resp = await self.handle_request(msg)
        msg.meta["status_code"] = resp.status
        msg.meta["url"] = str(resp.url)
        resp_content = resp.content
        if self.json:
            try:
                resp_content = json.loads(resp_content)
            except Exception:
                logger.exception(
                    "cannot json parse response from url %s (response=%r)",
                    self.generate_request_url(msg), resp_content)
        msg.payload = resp_content
        return msg



import csv
import io

from pypeman import nodes


class CSV2Python(nodes.BaseNode):
    """
    Convert CSV File to Python
    'headers' indicate if the csv file contains a first line with column titles
    'to_dict' indicates if module use csv.DictReader or csv.reader (
        to use csv.DictReader, headers must be set to True too)
    """
    def __init__(self, *args, filepath=None, to_dict=False, headers=False,
                 delimiter=',', quoting=None, encoding=None, **kwargs):
        self.delimiter = delimiter
        self.filepath = filepath
        self.headers = headers
        self.to_dict = to_dict and headers
        self.quoting = quoting or csv.QUOTE_NONE
        self.encoding = encoding or "utf-8"
        super().__init__(*args, **kwargs)

    def process(self, msg):
        data_to_return = []
        if self.filepath:
            fpath = nodes.callable_or_value(self.filepath, msg)
        else:
            fpath = msg.meta["filepath"]
        with open(fpath, "r", encoding=self.encoding) as fin:
            if self.to_dict:
                csvdata = csv.DictReader(fin, delimiter=self.delimiter, quoting=self.quoting)
            else:
                csvdata = csv.reader(fin, delimiter=self.delimiter, quoting=self.quoting)
                if self.headers:
                    csvdata.pop(0)
            for row in csvdata:
                data_to_return.append(row)
        msg.payload = data_to_return
        return msg


class CSVstr2Python(nodes.BaseNode):
    """
    Like CSV2Python but Convert CSV string to Python
    'headers' indicate if the csv file contains a first line with column titles
    'to_dict' indicates if module use csv.DictReader or csv.reader (
        to use csv.DictReader, headers must be set to True too)
    """
    def __init__(self, *args, filepath=None, to_dict=False, headers=False,
                 delimiter=',', quoting=None, encoding=None, **kwargs):
        self.delimiter = delimiter
        self.headers = headers
        self.to_dict = to_dict and headers
        self.quoting = quoting or csv.QUOTE_NONE
        self.encoding = encoding or "utf-8"
        super().__init__(*args, **kwargs)

    def process(self, msg):
        data_to_return = []
        csvfile = io.StringIO(msg.payload)
        if self.to_dict:
            csvdata = csv.DictReader(csvfile, delimiter=self.delimiter, quoting=self.quoting)
        else:
            csvdata = csv.reader(csvfile, delimiter=self.delimiter, quoting=self.quoting)
            if self.headers:
                csvdata.pop(0)
        for row in csvdata:
            data_to_return.append(row)
        msg.payload = data_to_return
        return msg


class Python2CSVstr(nodes.BaseNode):
    """
    Convert python list of dict (like json) to csv str
    """
    def __init__(self, *args, header=False, fieldnames=None,
                 delimiter=',', quoting=None, newline="", **kwargs):
        self.header = header
        self.delimiter = delimiter
        self.fieldnames = fieldnames
        self.quoting = quoting or csv.QUOTE_NONE
        self.newline = newline
        super().__init__(*args, **kwargs)

    def process(self, msg):
        data = msg.payload
        fieldnames = self.fieldnames or list(data[0].keys())
        output = io.StringIO(newline=self.newline)
        writer = csv.DictWriter(output, fieldnames, delimiter=self.delimiter, quoting=self.quoting)
        if self.header:
            writer.writeheader()
        for line in data:
            writer.writerow(line)
        output.seek(0)
        msg.payload = output.read()
        return msg



from pypeman.test import TearDownProjectTestCase as TestCase


class ImportTests(TestCase):
    def test_ftp_import(self):
        import pypeman.contrib.ftp  # noqa: F401

    def test_hl7_import(self):
        import pypeman.contrib.hl7  # noqa: F401

    def test_http_import(self):
        import pypeman.contrib.http  # noqa: F401

    def test_time_import(self):
        import pypeman.contrib.time  # noqa: F401

    def test_xml_import(self):
        import pypeman.contrib.xml  # noqa: F401



"""
Additional Nodes for working with the context
"""

import pypeman.nodes as nodes

from pypeman.nodes import NodeException


class CombineCtx(nodes.BaseNode):
    """
    creates payload dict with combination of different contexts
    """

    def __init__(self, ctx_names, meta_from=None, flatten=False,  *args, **kwargs):
        """
            param ctx_names: list of context names to save or dict with mapping
                   of context name to payload keys
            param: flatten: if true all ctx payloads (must be dicts) will be combined
                    in one dict
            param: meta_from: specifies which meta the resulting message should have
        """
        super().__init__(*args, **kwargs)
        if not isinstance(ctx_names, dict):
            if flatten:
                ctx_dst = len(ctx_names) * [None]
            else:
                ctx_dst = ctx_names
            ctx_names = dict(zip(ctx_names, ctx_dst))
        self.ctx_names = ctx_names
        if len(ctx_names) < 2:
            raise NodeException("must have at least two contexts for combining")
        if meta_from is None:
            meta_from = next(iter(ctx_names.keys()))
        self.meta_from = meta_from

    def process(self, msg):
        payload = {}
        ch_logger = self.channel.logger
        ch_logger.debug(
            "combine ctx_names = %s / meta from %r",
            repr(self.ctx_names),
            self.meta_from,
            )
        # TODO: can we just copy the meta if the payload changed?
        # TODO: If not, then rmv meta_from param and create new meta
        msg.meta = msg.ctx[self.meta_from]['meta']

        for ctx_name, dst in self.ctx_names.items():
            ch_logger.debug("combine ctx = %s -> %s", ctx_name, dst)
            if dst is None:
                ctx_payload = msg.ctx[ctx_name]['payload']
                ch_logger.debug(
                    "upd payload with ctx payload of %s ( %s )",
                    ctx_name, repr(ctx_payload))
                payload.update(ctx_payload)
            else:
                payload[dst] = msg.ctx[ctx_name]['payload']

        msg.payload = payload

        return msg



"""
This module contains all persistence related things.
"""

import asyncio
from concurrent.futures import ThreadPoolExecutor
from collections import defaultdict
import importlib
import logging

from sqlitedict import SqliteDict

from pypeman.conf import settings

SENTINEL = object()

logger = logging.getLogger(__name__)

_backend = None


async def get_backend(loop):
    """
    Return the configured backend instance.

    :param loop: Asyncio loop to use. Passed backend instance.
    """
    global _backend
    if not _backend:
        # Load backend on first use
        if not settings.PERSISTENCE_BACKEND:
            raise Exception("Persistence backend not configured.")

        module, _, class_ = settings.PERSISTENCE_BACKEND.rpartition('.')
        loaded_module = importlib.import_module(module)
        _backend = getattr(loaded_module, class_)(loop=loop, **settings.PERSISTENCE_CONFIG)

        await _backend.start()
    if _backend.loop != loop:
        logger.warning("Backend loop not the same as loop argument, changing it")
        _backend.loop = loop
    return _backend


class MemoryBackend():
    """
    Memory persistence backend.
    Only for testing purpose as this is not really persistent.
    """
    def __init__(self, loop=None):
        self.loop = loop
        self._data = defaultdict(dict)

    async def start(self):
        """
        Do nothing for this backend.
        """
        pass

    async def store(self, namespace, key, value):
        """ Store the value in a dict in memory

        :param namespace: Namespace for dict.
        :param key: Access key.
        :param value: Value to save.
        """
        self._data[namespace][key] = value

    async def get(self, namespace, key, default=SENTINEL):
        """
        Get the value from memory.

        :param namespace: Namespace for dict.
        :param key: Key to get.
        :param default: Default value if key missing.
        """
        if default is not SENTINEL:
            return self._data[namespace].get(key, default)
        else:
            return self._data[namespace][key]

    async def search_ids_by_value(self, namespace, value):
        found_ids = []
        for id, val in self._data[namespace].items():
            if val == value:
                found_ids.append(id)
        return found_ids

    async def get_num_entries(self, namespace):
        return len(self._data[namespace])


class SqliteBackend():
    """
    Sqlite persistence backend. Store data in an sqlite database with
    ACID garanties. Internally use a thread pool to execute database access.

    :param path: Path of sqlite file.
    :param thread_pool: If you want a specific thread_pool you can give one here.
    :param loop: Loop used for the executor.
    """
    def __init__(self, path, thread_pool=None, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.executor = thread_pool or ThreadPoolExecutor(max_workers=1)
        self.path = path

    async def start(self):
        """
        Do nothing for this backend.
        """
        pass

    def _sync_store(self, namespace, key, value):
        with SqliteDict(self.path, tablename=namespace) as pdict:
            pdict[key] = value
            pdict.commit()

    def _sync_get(self, namespace, key, default):
        with SqliteDict(self.path, tablename=namespace) as pdict:
            if default is not SENTINEL:
                return pdict.get(key, default)
            else:
                return pdict[key]

    def _search_ids_by_value(self, namespace, value):
        found_ids = []
        with SqliteDict(self.path, tablename=namespace) as pdict:
            for id, val in pdict.items():
                if val == value:
                    found_ids.append(id)
        return found_ids

    def _get_table_length(self, namespace):
        with SqliteDict(self.path, tablename=namespace) as pdict:
            return len(pdict)

    async def store(self, namespace, key, value):
        """ Store the value in a dict saved in sqlite db.

        :param namespace: Namespace table name.
        :param key: Access key.
        :param value: Value to save.
        """
        await self.loop.run_in_executor(self.executor, self._sync_store, namespace, key, value)

    async def get(self, namespace, key, default=SENTINEL):
        """
        Get the value from sqlite db.

        :param namespace: Namespace table name.
        :param key: Key to get.
        :param default: Default value if key missing.
        """
        return await self.loop.run_in_executor(self.executor, self._sync_get, namespace, key, default)

    async def get_num_entries(self, namespace):
        return await self.loop.run_in_executor(self.executor, self._get_table_length, namespace)

    async def search_ids_by_value(self, namespace, value):
        return await self.loop.run_in_executor(
            self.executor, self._search_ids_by_value, namespace, value)



import asyncio


class Sleeper:
    """
    Group of sleep calls allowing instant cancellation of all

    found at: https://stackoverflow.com/questions/37209864/interrupt-all-asyncio-sleep-currently-executing
    """

    def __init__(self, loop):
        self.loop = loop
        self.tasks = set()

    async def sleep(self, delay, result=None):
        """
        a sleep function, that can be interrupted
        """
        coro = asyncio.sleep(delay, result=result)
        task = asyncio.create_task(coro)
        self.tasks.add(task)
        try:
            return await task
        except asyncio.CancelledError:
            return result
        finally:
            self.tasks.remove(task)

    def cancel_all_helper(self):
        """
        Cancel all pending sleep tasks
        """
        cancelled = set()
        for task in self.tasks:
            if task.cancel():
                cancelled.add(task)
        return cancelled

    async def cancel_all(self):
        """
        Coroutine cancelling tasks
        """
        cancelled = self.cancel_all_helper()
        if self.tasks:
            await asyncio.wait(self.tasks)
        self.tasks -= cancelled
        return len(cancelled)



import asyncio
import logging
import sys

import warnings

import hl7

from pypeman import endpoints, channels, nodes, message
from pypeman.errors import PypemanParamError


logger = logging.getLogger(__name__)


class MLLPProtocol(asyncio.Protocol):
    """
    Minimal Lower-Layer Protocol (MLLP) takes the form:
        <VT>[HL7 Message]<FS><CR>

    References:
        - http://www.hl7standards.com/blog/2007/05/02/hl7-mlp-minimum-layer-protocol-defined/
        - http://www.hl7standards.com/blog/2007/02/01/ack-message-original-mode-acknowledgement/
    """

    def __init__(self, handler, loop=None):
        super().__init__()
        self._buffer = b''
        self.start_block = b'\x0b'  # <VT>, vertical tab
        self.end_block = b'\x1c'  # <FS>, file separator
        self.carriage_return = b'\x0d'  # <CR>, \r
        self.handler = handler
        self.loop = loop or asyncio.get_event_loop()

    def connection_made(self, transport):
        """
        Called when a connection is made.
        The argument is the transport representing the pipe connection.
        To receive data, wait for data_received() calls.
        When the connection is closed, connection_lost() is called.
        """
        self.transport = transport

    def process_response(self, future):
        self.writeMessage(future.result())

    def data_received(self, data):
        """
        Called when some data is received.
        The argument is a bytes object.
        """
        # try to find a complete message(s) in the combined the buffer and data
        messages = (self._buffer + data).split(self.end_block)
        # whatever is in the last chunk is an uncompleted message, so put back
        # into the buffer
        self._buffer = messages.pop(-1)

        for raw_message in messages:
            # strip the rest of the MLLP shell from the HL7 message
            raw_message = raw_message.strip(self.start_block + self.carriage_return)

            # only pass messages with data
            if len(raw_message) > 0:
                result = asyncio.create_task(self.handler(raw_message))
                result.add_done_callback(self.process_response)

    def writeMessage(self, message):
        # convert back to a byte string
        # wrap message in payload container
        self.transport.write(self.start_block + message + self.end_block + self.carriage_return)

    def connection_lost(self, exc):
        """
        Called when the connection is lost or closed.
        The argument is an exception object or None (the latter
        meaning a regular EOF is received or the connection was
        aborted or closed).
        """
        logger.info("MLLP: connection lost (exc=%s)", repr(exc))
        super().connection_lost(exc)


class MLLPEndpoint(endpoints.SocketEndpoint):

    def __init__(
            self,
            address=None, port=None,  # obsolete params
            encoding='utf-8',
            loop=None,
            host=None,
            sock=None,
            reuse_port=None,
            ):
        self.handlers = []
        self.address = address
        self.port = port
        self.loop = loop or asyncio.get_event_loop()
        self.handler = None
        if address or port:
            warnings.warn(
                "HTTPEndpoint 'address', 'adress' and 'port' params are deprecated. "
                "Replace it by 'host' or 'sock'", DeprecationWarning)
            if host or sock:
                raise PypemanParamError(
                    "Obsolete params ('adress', 'address', 'port') "
                    "can not be mixed with new params ('host', 'sock')")
            sock = address + ':' + str(port)
        if host and sock:
            raise PypemanParamError("There can only be one (parameter host or sock)")
        sock = sock or host

        if encoding != 'utf-8':
            warnings.warn("MLLPEndpoint 'encoding' parameters is deprecated", DeprecationWarning)
        self.encoding = encoding

        super().__init__(loop=loop, sock=sock, default_port='2100', reuse_port=reuse_port)

    def set_handler(self, handler):
        self.handler = handler

    async def start(self):
        self.make_socket()
        if self.handler:
            srv = await self.loop.create_server(
                protocol_factory=lambda: MLLPProtocol(self.handler, loop=self.loop),
                sock=self.sock_obj,
            )
            logger.debug("MLLP server started at %s", repr(self.sock))
            return srv
        else:
            logger.error("No MLLP handlers.")


class MLLPChannel(channels.BaseChannel):

    def __init__(self, *args, endpoint=None, encoding='utf-8', **kwargs):
        super().__init__(*args, **kwargs)
        if endpoint is None:
            raise TypeError('Missing "endpoint" argument')
        self.mllp_endpoint = endpoint

        if encoding is None:
            encoding = sys.getdefaultencoding()
        self.encoding = encoding

    async def start(self):
        await super().start()
        self.mllp_endpoint.set_handler(handler=self.handle_hl7_message)

    async def handle_hl7_message(self, hl7_message):
        content = hl7_message.decode(self.encoding)
        msg = message.Message(content_type='text/hl7', payload=content, meta={})
        try:
            await self.handle(msg)
            ack = hl7.parse(content, encoding=self.encoding)
            return str(ack.create_ack('AA')).encode(self.encoding)
        except channels.Dropped:
            ack = hl7.parse(content, encoding=self.encoding)
            return str(ack.create_ack('AA')).encode(self.encoding)
        except channels.Rejected:
            ack = hl7.parse(content, encoding=self.encoding)
            return str(ack.create_ack('AR')).encode(self.encoding)
        except Exception:
            logger.exception(
                "MLLPChannel %s raise an error, will send AE speed response", str(self))
            ack = hl7.parse(content, encoding=self.encoding)
            return str(ack.create_ack('AE')).encode(self.encoding)


class HL7ToPython(nodes.BaseNode):
    """ Convert hl7 payload to python struct."""

    def __init__(self, *args, encoding="utf-8", **kwargs):
        self.encoding = encoding
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = hl7.parse(msg.payload, encoding=self.encoding)
        msg.content_type = 'application/python'
        return msg


class PythonToHL7(nodes.BaseNode):
    """ Convert python payload to HL7. Must be HL7 structure."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = str(msg.payload)
        msg.content_type = 'text/hl7'
        return msg



from aiohttp import web

from pypeman.plugins.base import BasePlugin
from pypeman.plugins.remoteadmin import urls


class RemoteAdminPlugin(BasePlugin):

    def __init__(self, host="127.0.0.1", port=8091, url_prefix=""):
        super().__init__()
        self.app = web.Application()
        self.host = host
        self.port = port
        self.url_prefix = url_prefix
        self.runner = web.AppRunner(self.app)

    def ready(self):
        urls.init_urls(self.app, prefix=self.url_prefix)

    async def start(self):
        await self.runner.setup()
        site = web.TCPSite(self.runner, self.host, self.port)
        await site.start()

    async def stop(self):
        await self.runner.cleanup()



import asyncio

from pypeman import channels
from pypeman.channels import BaseChannel
from pypeman import nodes
from pypeman.test import TearDownProjectTestCase as TestCase
from pypeman.tests.common import generate_msg


class TstNode(nodes.BaseNode):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TestingTests(TestCase):
    def clean_loop(self):
        # Useful to execute future callbacks
        pending = asyncio.all_tasks(loop=self.loop)

        if pending:
            self.loop.run_until_complete(asyncio.gather(*pending))

    def start_channels(self):
        # Start channels
        for chan in channels.all_channels:
            self.loop.run_until_complete(chan.start())

    def setUp(self):
        # Create class event loop used for tests to avoid failing
        # previous tests to impact next test ? (Not sure)
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        # Remove thread event loop to be sure we are not using
        # another event loop somewhere
        asyncio.set_event_loop(None)

        # Avoid calling already tested channels
        channels.all_channels.clear()

    def tearDown(self):
        super().tearDown()
        self.clean_loop()

    def test_chan_process_in_test_mode(self):
        """ Whether BaseChannel handling with test mode is working """

        chan = BaseChannel(name="test_channel_test_1", loop=self.loop)

        n = TstNode(name="testme")
        chan.add(n)

        msg = generate_msg(message_content="X")

        # Launch channel processing
        self.start_channels()
        chan._reset_test()

        chan.handle_and_wait(msg)

        self.assertEqual(n.processed, 1, "Channel in test mode not working")
        self.assertTrue(hasattr(n, '_handle'), "Channel in test mode not working")
        self.assertEqual(chan.get_node("testme").last_input().payload, "X", "Last input broken")

    def test_chan_node_mocking_in_test_mode(self):
        """ Whether node mocking input and output is working """

        chan = BaseChannel(name="test_channel_test_2", loop=self.loop)

        n = TstNode(name="testme")
        chan.add(n)

        msg_x = generate_msg(message_content="X")
        msg_a = generate_msg(message_content="A")
        msg_b = generate_msg(message_content="B")
        msg_c = generate_msg(message_content="C")
        msg_d = generate_msg(message_content="D")

        def concat_e(msg):
            msg.payload += "E"
            return msg

        def concat_f(msg):
            msg.payload += "F"
            return msg

        # Launch channel processing
        self.start_channels()

        # Mock input
        chan._reset_test()
        chan.get_node("testme").mock(input=msg_a)

        ret = chan.handle_and_wait(msg_x)

        self.assertEqual(n.processed, 1, "Channel in test mode not working")
        self.assertEqual(ret.payload, "A", "Mocking input broken")
        self.assertEqual(chan.get_node("testme").last_input().payload, "X", "Last input broken")

        # Mock output
        chan._reset_test()
        chan.get_node("testme").mock(output=msg_b)
        ret = chan.handle_and_wait(msg_x)

        self.assertEqual(n.processed, 1, "Channel in test mode not working")
        self.assertEqual(ret.payload, "B", "Mocking input broken")
        self.assertEqual(chan.get_node("testme").last_input().payload, "X", "Last input broken")

        # Mock both
        chan._reset_test()
        chan.get_node("testme").mock(input=msg_c, output=msg_d)
        ret = chan.handle_and_wait(msg_x)

        self.assertEqual(n.processed, 1, "Channel in test mode not working")
        self.assertEqual(ret.payload, "D", "Mocking both input and output broken")

        # Mock both functions
        chan._reset_test()
        chan.get_node("testme").mock(input=concat_e, output=concat_f)
        ret = chan.handle_and_wait(msg_x)

        self.assertEqual(n.processed, 1, "Channel in test mode not working")
        self.assertEqual(ret.payload, "XEF", "Mocking with function broken")



import asyncio
import cmd
import functools
import json
import logging
import re
import sys

from datetime import datetime
# TODO use readline for history ?
# import readline

import contextlib
import websockets

from aiohttp import web
from io import StringIO

from jsonrpcserver import async_dispatch
from jsonrpcserver import method
from jsonrpcclient import parse_json
from jsonrpcclient import request_json


from pypeman.conf import settings
from pypeman import channels
from pypeman import message

logger = logging.getLogger(__name__)


@contextlib.contextmanager
def stdoutIO(stdout=None):
    old = sys.stdout
    if stdout is None:
        stdout = StringIO()
    sys.stdout = stdout
    yield stdout
    sys.stdout = old


class RemoteAdminServer():
    """
    Expose json/rpc function to a client by a websocket.
    """

    def __init__(self, loop=None, host='localhost', port='8091', ssl=None, url=None):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.loop = loop or asyncio.get_event_loop()
        self.ctx = {}

    def get_channel(self, name):
        """
        return channel by is name.all_channels
        """
        for chan in channels.all_channels:
            if chan.name == name:
                return chan
        return None

    async def start(self):
        """ Start remote admin server """

        start_server = websockets.serve(
            self.command,
            host=self.host,
            port=self.port,
            ssl=self.ssl,
            loop=self.loop
        )
        await start_server

    async def command(self, websocket, path):
        """
        Generic function to handle a command from client.
        """
        request = await websocket.recv()
        response = await async_dispatch(request, context=self)
        if response.wanted:
            await websocket.send(str(response))

    @method
    async def exec(self, command):
        """
        Execute a python command on this instance and
        return the stdout result.

        :param command: The python command to execute. Can be multiline.
        :returns: Command stdout result.
        """
        # TODO may cause problem on multi thread access to stdout
        # as we highjack sys.stdout
        with stdoutIO() as out:
            exec(command, globals())

        return out.getvalue()

    @method
    async def channels(self):
        """
        Return a list of available channels.
        """
        chans = []
        for chan in channels.all_channels:
            if not chan.parent:
                chan_dict = chan.to_dict()
                chan_dict['subchannels'] = chan.subchannels()

                chans.append(chan_dict)

        return chans

    @method
    async def start_channel(self, channel):
        """
        Start the specified channel

        :params channel: The channel name to start.
        """
        chan = self.get_channel(channel)
        await chan.start()
        return {
            'name': chan.name,
            'status': channels.BaseChannel.status_id_to_str(chan.status)
        }

    @method
    async def stop_channel(self, channel):
        """
        Stop the specified channel

        :params channel: The channel name to stop.
        """
        chan = self.get_channel(channel)
        await chan.stop()
        return {
            'name': chan.name,
            'status': channels.BaseChannel.status_id_to_str(chan.status)
        }

    @method
    async def list_msgs(self, channel, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                        text=None, rtext=None):
        """
        List first `count` messages from message store of specified channel.

        :params channel: The channel name.
        """
        chan = self.get_channel(channel)

        messages = await chan.message_store.search(
            start=start, count=count, order_by=order_by, start_dt=start_dt, end_dt=end_dt,
            text=text, rtext=rtext) or []

        for res in messages:
            timestamp = res['timestamp']
            if isinstance(timestamp, datetime):
                timestamp = datetime.timestamp(timestamp)
            res['timestamp'] = timestamp
            res['id'] = res['id']
            res['state'] = res["state"]
            if "message" in res:
                res.pop("message")

        return {'messages': messages, 'total': await chan.message_store.total()}

    @method
    async def replay_msg(self, channel, msg_id):
        """
        Replay messages from message store.

        :params channel: The channel name.
        :params msg_ids: The message ids list to replay.
        """
        chan = self.get_channel(channel)
        try:
            msg_res = await chan.replay(msg_id)
            result = msg_res.to_dict()
        except Exception as exc:
            result = {'error': str(exc)}
        return result

    @method
    async def view_msg(self, channel, msg_id):
        """
        Permit to get the content of a message

        :params channel: The channel name.
        :params msg_ids: The message ids list to replay.
        """
        chan = self.get_channel(channel)
        try:
            msg_res = await chan.message_store.get_msg_content(msg_id)
            result = msg_res.to_dict()
        except Exception as exc:
            result = {'error': str(exc)}

        return result

    @method
    async def preview_msg(self, channel, msg_id):
        """
        Permits to get the 1000 chars of a message payload

        :params channel: The channel name.
        :params msg_ids: The message ids list to replay.
        """
        chan = self.get_channel(channel)
        try:
            msg_res = await chan.message_store.get_preview_str(msg_id)
            result = msg_res.to_dict()
        except Exception as exc:
            result = {'error': str(exc)}
        return result

    @method
    async def push_msg(self, channel, text):
        """
        Push a message in the channel.

        :params channel: The channel name.
        :params msg_ids: The text added to the payload.
        """
        chan = self.get_channel(channel)
        msg = message.Message(payload=text)
        result = await chan.handle(msg)
        return result.to_dict()


class RemoteAdminClient():
    """
    Remote admin client. To be use by ipython shell or pypeman shell.

    :params url: Pypeman Websocket url.
    """
    def __init__(self, loop=None, url='ws://localhost:8091'):
        self.url = url
        self.loop = loop or asyncio.get_event_loop()

    def init(self):
        pass

    def send_command(self, command, args=None):
        """
        Send a command to remote instance
        """
        if args is None:
            args = []
        if not isinstance(args, list):
            args = [args]
        return self.loop.run_until_complete(self._send_command(command, args))

    async def _send_command(self, command, args):
        """
        Asynchronous version of command sending
        """
        async with websockets.connect(self.url) as ws:
            await ws.send(request_json(command, args))
            response = await ws.recv()
            parsed_response = parse_json(response)
            return parsed_response.result

    def exec(self, command):
        """
        Execute any valid python code on remote instance
        and return stdout result.
        """
        result = self.send_command('exec', [command])
        return result

    def channels(self):
        """
        Return a list of available channels on remote instance.
        """
        return self.send_command('channels')

    def start(self, channel):
        """
        Start the specified channel on remote instance.

        :params channel: The channel name.
        """
        return self.send_command('start_channel', [channel])

    def stop(self, channel):
        """
        Stop the specified channel on remote instance.

        :params channel: The channel name.
        """
        return self.send_command('stop_channel', [channel])

    def list_msgs(self, channel, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                  text=None, rtext=None):
        """
        List first 10 messages on specified channel from remote instance.

        :params channel: The channel name.
        :params start: Start index of listing.
        :params count: Count from index.
        :params start_dt: (optional) start datetime filter (isoformat)
        :params end_dt: (optional) start datetime filter (isoformat)
        :params text: (optional) text to search in message
        :params rtext: (optional) regex to search in message
        :params order_by: Message order. only 'timestamp' and '-timestamp' handled for now.
        :returns: list of message with status.
        """
        list_msg_args = [channel, start, count, order_by, start_dt, end_dt, text, rtext]
        result = self.send_command('list_msgs', list_msg_args)

        for m in result['messages']:
            if "message" in m:
                m.pop("message")

        return result

    def replay_msg(self, channel, msg_id):
        """
        Replay specified message from id list of specified channel on remote instance.

        :params channel: The channel name.
        :params msg_id: Message id to replay
        :returns: List of result for each message. Result
            can be {'error': <msg_error>} for one id if error
            occurs.
        """
        request_result = self.send_command('replay_msg', [channel, msg_id])

        if 'error' not in request_result:
            result = message.Message.from_dict(request_result)
        else:
            result = request_result

        return result

    def view_msg(self, channel, msg_id):
        """
        View content of the specified message from id of specified
        channel on remote instance.

        :params channel: The channel name.
        :params msg_ids: Message id list to replay
        :returns: List of result for each message. Result
            can be {'error': <msg_error>} for one id if error
            occurs.
        """
        request_result = self.send_command('view_msg', [channel, msg_id])

        if 'error' not in request_result:
            result = message.Message.from_dict(request_result)
        else:
            result = request_result

        return result

    def preview_msg(self, channel, msg_id):
        """
        Preview content of the specified message from id of specified
        channel on remote instance.

        :params channel: The channel name.
        :params msg_ids: Message id list to replay
        :returns: List of result for each message. Result
            can be {'error': <msg_error>} for one id if error
            occurs.
        """
        request_result = self.send_command('preview_msg', [channel, msg_id])

        if 'error' not in request_result:
            result = message.Message.from_dict(request_result)
        else:
            result = request_result

        return result

    def push_msg(self, channel, text):
        """
        Push a new message from text param to the channel.

        :params channel: The channel name.
        :params text: This text will be the payload of the message.
        """
        msg_dict = self.send_command('push_msg', [channel, text])
        return message.Message.from_dict(msg_dict)


def _with_current_channel(func):
    """
    Decorator to select channel for command.
    """
    @functools.wraps(func)
    def wrapper(self, *arg, **kwargs):
        if self.current_channel:
            return func(self, self.current_channel, *arg, **kwargs)
        else:
            print("You have to select a channel prior using this command")
            return None

    return wrapper


class PypemanShell(cmd.Cmd):
    intro = 'Welcome to the pypeman shell. Type help or ? to list commands.\n'
    prompt = 'pypeman > '
    use_rawinput = False

    def __init__(self, url):
        super().__init__()
        self.current_channel = None
        self.client = RemoteAdminClient(url=url)
        self.client.init()

    def do_channels(self, arg):
        "List avaible channels"
        result = self.client.channels()
        print("\nChannel list:")
        for idx, channel in enumerate(result):
            print("{idx}) {name} ({status})".format(idx=idx, **channel))

        print("")
        self.current_channel = None
        self.prompt = "pypeman > "

    def do_stop(self, arg):
        "Stop a channel by his name"
        result = self.client.stop(arg)
        print(result)

    def do_start(self, arg):
        "Start a channel by his name"
        result = self.client.start(arg)
        print(result)

    def do_select(self, arg):
        """
        Select a channel by is name.
        Mandatory for channel oriented command
        """
        self.current_channel = arg
        self.prompt = "pypeman(%s) > " % arg

    def get_current_channel(self):
        if self.current_channel:
            return self.current_channel
        else:
            print("You have to select a channel before for this command")
            return None

    @_with_current_channel
    def do_list(self, channel, arg):
        """
        List messages of selected channel. You can specify start, end and order_by arguments
        Optional args:
            - to filter messages, you can pass start_dt and end_dt (isoformat datetimes)
            to filter messages
            - to filter messages, you can pass text and rtext (string between double quotes)
            to filter messages
        """
        dquote_args_regex = r'\w+=".*?"'
        dquote_args = re.findall(dquote_args_regex, arg)
        for st_arg in dquote_args:
            arg = arg.replace(st_arg, "")
        dquote_args = [dqarg.replace('"', "") for dqarg in dquote_args]
        args = arg.split()
        if dquote_args:
            args.extend(dquote_args)
        start, end, order_by = 0, 100, '-timestamp'
        start_dt = None
        end_dt = None
        text = None
        rtext = None

        args_copy = [i for i in args]
        # Parsing of naming args
        for arg in args_copy:
            if isinstance(arg, str):
                if arg.startswith("start_dt="):
                    start_dt = arg.split("=")[1]
                    args.remove(arg)
                if arg.startswith("end_dt="):
                    end_dt = arg.split("=")[1]
                    args.remove(arg)
                if arg.startswith("text="):
                    text = arg.split("=", 1)[1]
                    args.remove(arg)
                if arg.startswith("rtext="):
                    rtext = arg.split("=", 1)[1]
                    args.remove(arg)

        # Parsing of common args
        if args:
            start = int(args[0])
        if len(args) > 1:
            end = int(args[1])
        if len(args) > 2:
            order_by = args[2]
        result = self.client.list_msgs(
            channel, start, end, order_by, start_dt=start_dt, end_dt=end_dt,
            text=text, rtext=rtext)

        if not result['total']:
            print('No message yet.')

        for res in result['messages']:
            print(res['timestamp'], res['id'], res['state'])

    @_with_current_channel
    def do_replay(self, channel, arg):
        "Replay a message list by their ids"
        msg_ids = str(arg).split()
        for msg_id in msg_ids:
            msg = self.client.replay_msg(channel, msg_id)
            print("Result message for replaying message %s:" % msg_id)
            if isinstance(msg, message.Message):
                print(msg.to_print())
            else:
                print('Error on msg: %s - %s' % (msg_id, msg['error']))

    @_with_current_channel
    def do_view(self, channel, arg):
        "View content of a message list by their ids"
        msg_ids = arg.split()
        for msg_id in msg_ids:
            msg = self.client.view_msg(channel, msg_id)
            print("View of msg %s:" % msg_id)
            if isinstance(msg, message.Message):
                print(msg.to_print())
            else:
                print('Error on msg: %s - %s' % (msg_id, msg['error']))

    @_with_current_channel
    def do_preview(self, channel, arg):
        "Preview content of a message list by their ids"
        msg_ids = arg.split()
        for msg_id in msg_ids:
            msg = self.client.preview_msg(channel, msg_id)
            print("Preview of message %s:" % msg_id)
            if isinstance(msg, message.Message):
                print(msg.to_print())
            else:
                print('Error on msg: %s - %s' % (msg_id, msg['error']))

    @_with_current_channel
    def do_push(self, channel, arg):
        "Inject message with text as payload for selected channel"
        result = self.client.push_msg(channel, arg)
        print("Result message:")
        print(result.to_print())

    def do_exit(self, arg):
        "Exit program"
        sys.exit()


class WebAdmin():
    def __init__(self, loop, host, port, ssl):
        self.host = host
        self.port = port
        self.ssl = ssl
        self.loop = loop

    async def start(self):

        # client_dir = os.path.join(os.path.dirname(os.path.join(__file__)), 'client/dist')
        app = web.Application()
        app.router.add_get(
            '/configs.js',
            self.handle_config
        )

        # redirect to index.html
        app.router.add_get('/', self.redirect_to_index)

        # app.router.add_static(
        #     '/',
        #     path=os.path.join(client_dir),
        #     name='static'
        # )

        await self.loop.create_server(
            protocol_factory=app.make_handler(),
            host=self.host,
            port=self.port,
            ssl=self.ssl
        )

    async def redirect_to_index(self, request):
        """ redirect to index.html"""
        return web.HTTPFound("index.html")

    async def handle_config(self, request):
        conf = settings.REMOTE_ADMIN_WEBSOCKET_CONFIG
        server_url = conf['url'] or "ws{is_secure}://{host}:{port}".format(
            is_secure='s' if conf['ssl'] else '',
            **conf
        )

        conf = {
            'serverConfig': server_url
        }
        # TODO Not really cool but works.
        resp = """window.configs = {};""".format(json.dumps(conf))

        return web.Response(text=resp)



"""
dispatcher command for pypeman tools
"""
import argparse
import importlib
import sys

from functools import partial


commands = [
    ("view_store", "read and search in file a pypeman store"),
    ("send_from_store", "send entry from store"),
]


def list_commands():
    """
    list command and it's short description
    """
    print("List of commands:")
    for cmd, descr in commands:
        print(f"{cmd}: {descr}")


def imp_mod_and_run(cmd, args):
    """
    import module and run its main func
    """
    mod = importlib.import_module(f"pypeman.tools.{cmd}")
    sys.argv[1:] = args
    mod.main()


def mk_parser():
    """
    create argument parser
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-l", "--list-commands", help="list commands with small description", action="store_true")
    subparsers = parser.add_subparsers(dest="command", help="sub command to call")
    for cmd, descr in commands:
        sub_prs = subparsers.add_parser(cmd, description=descr, add_help=False)
        sub_prs.set_defaults(func=partial(imp_mod_and_run, cmd))
    return parser


def main():
    parser = mk_parser()
    options, args = parser.parse_known_args()
    if options.list_commands:
        list_commands()
        return
    if not options.command:
        parser.print_help()
        return
    options.func(args)


if __name__ == "__main__":
    main()



import asyncio
from unittest import mock

import pytest

from pypeman import channels
from pypeman import nodes
from pypeman.errors import PypemanError
from pypeman.graph import mk_graph
from pypeman.graph import wait_for_loop
from pypeman.tests.pytest_helpers import clear_graph  # noqa: F401


class FakeChannel:
    def __init__(self, loop=None):
        if loop:
            self.loop

    def __repr__(self):
        return "FakeChannel(loop=" + repr(getattr(self, "loop", "not set")) + ")"


def mk_fake_all_channels(t=0):
    """
    helper to moke pypeman.channels.all_channels

    :param t: time after which channel.loop should be
              set
    """
    channel = FakeChannel(loop=None)
    all_channels = [channel]
    if t < 0:
        return all_channels
    if t == 0:
        channel.loop = True
        return all_channels
    # TODO: start thread to set channel.loop from 0 to 1
    return all_channels


@mock.patch('pypeman.channels.all_channels', mk_fake_all_channels(t=-1))
def test_have_no_loop():
    with pytest.raises(PypemanError):
        wait_for_loop(tmax=0.1)


@mock.patch('pypeman.channels.all_channels', mk_fake_all_channels(t=0))
def test_have_loop():
    loop = wait_for_loop(tmax=0.1)
    print("loop =", loop)
    assert loop, "loop should be set, but is %s" % loop


@pytest.mark.usefixtures("clear_graph")
def test_show_graph():
    """ is a graph generated """
    loop = asyncio.new_event_loop()
    print("loop =", loop)
    names = (f"n{cnt}" for cnt in range(10))
    chan1 = channels.BaseChannel(name=next(names), loop=loop)
    chan1.add(nodes.BaseNode(name=next(names)))
    chan1.add(nodes.BaseNode(name=next(names)))
    subchan = chan1.when(lambda v: True, name=next(names))
    subchan.add(nodes.BaseNode(name=next(names)))
    rslt = list(mk_graph())
    as_str = "\n".join(rslt)
    print("RSLT:", as_str)
    print(repr(as_str))
    assert len(rslt) == 8
    assert as_str == (
        "BaseChannel\n|-n1\n|-n2\n|?\\ (n0.n3)\n|"
        "  |-n4\n|  -> Out\n|-> out\n")



import asyncio

from pypeman.helpers.aio_compat import awaitify


class Event:
    """
    Asyncio Event class.
    """

    def __init__(self):
        self.handlers = set()

    def add_handler(self, handler):
        """
        Add a new handler for this event.
        """
        self.handlers.add(handler)
        return self

    def remove_handler(self, handler):
        """
        Remove a previously defined handler for this event.
        """
        try:
            self.handlers.remove(handler)
        except Exception:
            raise ValueError("Handler is not handling this event, so cannot unhandle it.")
        return self

    def receiver(self, handler):
        """
        Function decorator to add an handler.
        """
        self.add_handler(handler)
        return handler

    async def fire(self, *args, **kargs):
        """
        Fire current event. All handler are going to be executed.
        """
        for handler in self.handlers:
            if not asyncio.iscoroutinefunction(handler):
                handler = awaitify(handler)

            await handler(*args, **kargs)

    def getHandlerCount(self):
        """
        Return declared handler count.
        """
        return len(self.handlers)

    __iadd__ = add_handler
    __isub__ = remove_handler
    __call__ = fire
    __len__ = getHandlerCount


channel_change_state = Event()



#!/usr/bin/env python

""" Helpers for testing
- helpers for testing pypeman itself
- helpers for testing pypeman projects
"""

from unittest import TestCase

import asyncio
import logging

from pypeman import nodes
from pypeman import channels


logger = logging.getLogger(__name__)
# TODO implement settings override
# TODO implement MessageStoreMock


class PypeTestCase(TestCase):
    """ Test Case to be used for testing pypeman projects

        This test case ensures, that
        event loops and other asyncio and pypeman specifics
        are correctly cleaned up / set up between unit test
        runs.

        This class is necessary as asyncio and pypeman have
        some global persistent objects like
        asyncio default loop
        channels wait_subchans param (permits to raise exc in current loop)
        pypeman.nodes.all_nodes / pypeman.channels.all_channels, ...

        Anybody using unittest.TestCase based tests for a pypeman
        project should use this class instead.
    """

    loop = None
    wait_subchans = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.addCleanup(self.cleanLoop)

    def setUp(self):
        if self.loop:
            raise Exception("a loop currently exists at setuP, not normal")
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        # Start channels
        for chan in channels.all_channels:
            chan.loop = self.loop
            chan.wait_subchans = self.wait_subchans
            self.loop.run_until_complete(chan.start())
            chan._reset_test()

    def cleanLoop(self):
        for chan in channels.all_channels:
            self.loop.run_until_complete(chan.stop())
        self.loop.close()
        self.loop.stop()
        self.loop = None
        asyncio.set_event_loop(None)

    def finish_all_tasks(self):
        """
        You can use this function if you have some subchannel in you channel
        , don't have wait_subchans param set to True and anyway want to see the
        final result by processing all remaining tasks.

        :return: A list of raised exceptions during task execution.
        """
        if self.wait_subchans:
            logger.warning(
                "PypeTestCase.finish_all_tasks called when wait_subchans is set..."
                "useless, there won't be any pending tasks"
            )
        raised_exceptions = []
        pending = asyncio.all_tasks(loop=self.loop)

        for task in pending:
            try:
                self.loop.run_until_complete(task)
            except Exception as exc: # noqa
                raised_exceptions.append(exc)
        return raised_exceptions

    def get_channel(self, name):
        """
        Return a channel by is name. Remember to prepend with parent channel
        name for subchannel.

        :return: Channel instance corresponding to `name`
            or None if channel not found.
        """
        for chan in channels.all_channels:
            if chan.name == name:
                chan._reset_test()
                return chan
        raise NameError("Channel '%s' doesn't exist" % name)

    def set_loop_to_debug(self):
        """
        :return:
        """
        self.loop.set_debug(True)


class TearDownProjectTestCase(TestCase):
    """ unittest for tests creating a project

    this testcase ensures, that at the end of the test the project is
    removed
    """

    def tearDown(self):
        clear_pypeman_project()


def clear_pypeman_project():
    """ clear a pypeman project. should be used whenever tests don't need

        a certain pypeman graph anymore and one wants to be sure that
        residuals are cleaned away
    """

    nodes.reset_pypeman_nodes()



"""
Default configuration file.
This settings should be redifined.
"""

DEBUG = False    # bool. can be set by env var PYPEMAN_DEBUG (0|1|true|false) or pypeman cmd args
TESTING = False  # bool. can be set by env var PYPEMAN_TESTING (0|1|true|false) pypeman cmd args

PROJECT_MODULE = "project"  # name of module containing the project

DEBUG_PARAMS = dict(
    slow_callback_duration=0.1
)

REMOTE_ADMIN_WEBSOCKET_CONFIG = {
    'host': 'localhost',
    'port': '8091',
    'ssl': None,
    'url': None,  # must be set when behind a reverse proxy
}

REMOTE_ADMIN_WEB_CONFIG = {
    'host': 'localhost',
    'port': '8090',
    'ssl': None,
}

HTTP_ENDPOINT_CONFIG = ['0.0.0.0', '8080']

PERSISTENCE_BACKEND = None
PERSISTENCE_CONFIG = {}

PLUGINS = []

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(name)s %(module)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        }
    },

    'loggers': {
        '': {
            'level': 'INFO',
            'handlers': ['console'],
        },
        'jsonrpcclient': {
            'level': 'WARNING',
            'propagate': False,
        },
        'jsonrpcserver': {
            'level': 'WARNING',
            'propagate': False,
        }
    }
}



"""
tests for pypeman.contrib.ctx
"""

import asyncio

import pytest

from pypeman.contrib.ctx import CombineCtx
from pypeman.nodes import NodeException
from pypeman.tests.pytest_helpers import clear_graph  # noqa: F401

from pypeman.tests.common import generate_msg
# TODO: might refactor to another file?
from pypeman.tests.test_nodes import FakeChannel


def mk_msgs_w_ctx(*ctx_ids):
    """
    helper to create a msg with a few contexts
    """
    msg = generate_msg(
        message_content="test",
        message_meta=dict(entry1="meta1"),
        )
    ctx_msgs = []
    for ctx_id in ctx_ids:
        meta = dict(entry1="meta_%s" % ctx_id)
        ctx_msg = generate_msg(
            message_content={"val_%s" % ctx_id: "data_%s" % ctx_id},
            message_meta=meta,
            )
        msg.add_context(ctx_id, ctx_msg)
        ctx_msgs.append(ctx_msg)

    return msg, ctx_msgs


@pytest.mark.usefixtures("clear_graph")
def test_combine_ctx_not_two_names(event_loop):
    with pytest.raises(NodeException):
        CombineCtx([], name="ctx1")
    with pytest.raises(NodeException):
        CombineCtx(["a"], name="ctx2")


@pytest.mark.usefixtures("clear_graph")
def test_combine_ctx_2_names(event_loop):
    loop = event_loop
    asyncio.set_event_loop(loop)
    # nut == Node Under Test
    nut = CombineCtx(["a", "b"], name="ctx1")
    nut.channel = FakeChannel(loop)
    msg, ctx_msgs = mk_msgs_w_ctx("a", "b")

    rslt = loop.run_until_complete(nut.handle(msg))
    assert rslt.payload["a"] == ctx_msgs[0].payload
    assert rslt.payload["b"] == ctx_msgs[1].payload
    assert rslt.meta == ctx_msgs[0].meta


@pytest.mark.usefixtures("clear_graph")
def test_combine_ctx_2_names_w_meta(event_loop):
    loop = event_loop
    asyncio.set_event_loop(loop)
    # nut == Node Under Test
    nut = CombineCtx(["a", "b"], meta_from="b", name="ctx1")
    nut.channel = FakeChannel(loop)
    msg, ctx_msgs = mk_msgs_w_ctx("a", "b")

    rslt = loop.run_until_complete(nut.handle(msg))
    assert rslt.payload["a"] == ctx_msgs[0].payload
    assert rslt.payload["b"] == ctx_msgs[1].payload
    assert rslt.meta == ctx_msgs[1].meta


@pytest.mark.usefixtures("clear_graph")
def test_combine_ctx_2_names_flat(event_loop):
    loop = event_loop
    asyncio.set_event_loop(loop)
    # nut == Node Under Test
    nut = CombineCtx(["a", "b"], name="ctx1", flatten=True)
    nut.channel = FakeChannel(loop)
    msg, ctx_msgs = mk_msgs_w_ctx("a", "b")

    rslt = loop.run_until_complete(nut.handle(msg))
    exp_payload = dict(ctx_msgs[0].payload)
    exp_payload.update(ctx_msgs[1].payload)
    assert rslt.payload == exp_payload
    assert rslt.meta == ctx_msgs[0].meta



import json
import logging

from aiohttp import web
from jsonrpcserver.response import SuccessResponse

from pypeman import channels

logger = logging.getLogger(__name__)


def get_channel(name):
    """
    returns channel by name
    """
    for chan in channels.all_channels:
        if chan.name == name:
            return chan
    return None


async def list_channels(request, ws=None):
    """
    Returns a list of available channels.
    """

    chans = []
    for chan in channels.all_channels:
        if not chan.parent:
            chan_dict = chan.to_dict()
            chan_dict['subchannels'] = chan.subchannels()

            chans.append(chan_dict)

    if ws is not None:
        await ws.send_jsonrpcresp(chans)
        return ws
    return web.json_response(chans)


async def start_channel(request, ws=None):
    """
    Start the specified channel

    :params channelname: The channel name to start.
    """
    channelname = request.match_info['channelname']
    chan = get_channel(channelname)
    await chan.start()
    resp_dict = {
        'name': chan.name,
        'status': channels.BaseChannel.status_id_to_str(chan.status)
    }
    if ws is not None:
        await ws.send_jsonrpcresp(resp_dict)
        return ws
    return web.json_response(resp_dict)


async def stop_channel(request, ws=None):
    """
    Stop the specified channel

    :params channelname: The channel name to stop.
    """
    channelname = request.match_info['channelname']

    chan = get_channel(channelname)
    await chan.stop()

    resp_dict = {
        'name': chan.name,
        'status': channels.BaseChannel.status_id_to_str(chan.status)
    }
    if ws is not None:
        await ws.send_jsonrpcresp(resp_dict)
        return ws
    return web.json_response(resp_dict)


async def list_msgs(request, ws=None):
    """
    List first `count` message infos from message store of specified channel.
    Return list of dicts with the id, the status and the timestamp of the message
    without the content.

    :params channelname: The channel name.

    :queryparams:
        start (int, default=0): the start indexof msgs to list
        count (count, default=10): The maximum returned msgs
        order_by (str, default="timestamp"): the message attribute to use for sorting
        start_dt (str): iso datetime string to use for filter messages
        end_dt (str): iso datetime string to use for filter messages
        text (str): search this text in messages and return only messages that contains
                    this text
        rtext (str): same as 'text' param but for regex
    """

    channelname = request.match_info['channelname']

    chan = get_channel(channelname)

    args = request.rel_url.query
    start = int(args.get("start", 0))
    count = int(args.get("count", 10))
    order_by = args.get("order_by", "-timestamp")
    start_dt = args.get("start_dt", None)
    end_dt = args.get("end_dt", None)
    text = args.get("text", None)
    rtext = args.get("rtext", None)
    messages = await chan.message_store.search(
        start=start, count=count, order_by=order_by, start_dt=start_dt, end_dt=end_dt,
        text=text, rtext=rtext) or []

    for res in messages:
        res["id"] = res["id"]
        res["state"] = res.get("state", "UNKNOWN")
        res['timestamp'] = res['message'].timestamp_str()
        if "message" in res:
            res.pop("message")

    resp_dict = {'messages': messages, 'total': await chan.message_store.total()}
    if ws is not None:
        await ws.send_jsonrpcresp(resp_dict)
        return ws
    return web.json_response(resp_dict)


async def replay_msg(request, ws=None):
    """
    Replay message from message store.

    :params channel: The channel name.
    :params msg_id: The message id to replay.

    :queryparams:
        encode_payload (Bool, default=False): Force pickling and encoding the payload or not
    """
    channelname = request.match_info['channelname']
    message_id = request.match_info['message_id']
    args = request.rel_url.query
    encode_payload = args.get("encode_payload", False)
    chan = get_channel(channelname)
    try:
        msg_res = await chan.replay(message_id)
        result = msg_res.to_dict(encode_payload=encode_payload)
    except IndexError:
        message = f"Cannot replay msg, id {message_id} probably doesn't exists"
        logger.error(message)
        result = {'error': message}
    except Exception as exc:
        logger.exception(f"Cannot replay msg {message_id}")
        result = {'error': str(exc)}

    if ws is not None:
        await ws.send_jsonrpcresp(result)
        return ws
    return web.json_response(result)


async def view_msg(request, ws=None):
    """
    Permit to get the content of a message

    :params channelname: The channel name.
    :params message_id: The message id to view

    :queryparams:
        encode_payload (Bool, default=False): Force pickling and encoding the payload or not
    """

    channelname = request.match_info['channelname']
    message_id = request.match_info['message_id']

    args = request.rel_url.query
    encode_payload = args.get("encode_payload", False)

    chan = get_channel(channelname)

    try:
        msg_res = await chan.message_store.get_msg_content(message_id)
        result = msg_res.to_dict(encode_payload=encode_payload)
    except IndexError:
        message = f"Cannot view msg, id {message_id} probably doesn't exists"
        logger.error(message)
        result = {'error': message}
    except Exception as exc:
        logger.exception(f"Cannot view msg {message_id}")
        result = {'error': str(exc)}

    if ws is not None:
        await ws.send_jsonrpcresp(result)
        return ws
    return web.json_response(result)


async def preview_msg(request, ws=None):
    """
    Permits to get the first 1000 chars of a message payload

    :params channelname: The channel name.
    :params message_id: The message id to preview

    :queryparams:
        encode_payload (Bool, default=False): Force pickling and encoding the payload or not
    """
    channelname = request.match_info['channelname']
    message_id = request.match_info['message_id']

    args = request.rel_url.query
    encode_payload = args.get("encode_payload", False)

    chan = get_channel(channelname)
    try:
        msg_res = await chan.message_store.get_preview_str(message_id)
        result = msg_res.to_dict(encode_payload=encode_payload)
    except IndexError:
        message = f"Cannot preview msg, id {message_id} probably doesn't exists"
        logger.error(message)
        result = {'error': message}
    except Exception as exc:
        logger.exception(f"Cannot preview msg {message_id}")
        result = {'error': str(exc)}

    if ws is not None:
        await ws.send_jsonrpcresp(result)
        return ws
    return web.json_response(result)


class RPCWebSocketResponse(web.WebSocketResponse):
    """
    Mocked aiohttp.web.WebSocketResponse to return JSOn RPC responses
    Workaround to have a backport compatibility with old json rpc client
    """

    def set_rpc_attrs(self, request_data):
        self.rpc_data = request_data

    async def send_jsonrpcresp(self, message):
        message = SuccessResponse(message, id=self.rpc_data["id"])
        await super().send_str(str(message))


async def backport_old_client(request):
    ws = RPCWebSocketResponse()
    logger.debug("Receiving a request from the shell client (%r)", vars(request))
    await ws.prepare(request)
    async for msg in ws:
        try:
            cmd_data = json.loads(msg.data)
        except Exception:
            await ws.send_str(f"cannot parse ws json data ({msg.data})")
            return ws
        ws.set_rpc_attrs(cmd_data)
        cmd_method = cmd_data.pop("method")
        params = cmd_data.get("params", [None])
        channelname = params[0]
        if channelname:
            request.match_info["channelname"] = channelname

        if cmd_method == "channels":
            await list_channels(request, ws=ws)
        elif cmd_method == "preview_msg":
            message_id = params[1]
            request.match_info["message_id"] = message_id
            query_url = request.rel_url.with_query({"encode_payload": "True"})
            new_req = request.clone(rel_url=query_url)
            await preview_msg(request=new_req, ws=ws)
        elif cmd_method == "view_msg":
            message_id = params[1]
            request.match_info["message_id"] = message_id
            query_url = request.rel_url.with_query({"encode_payload": "True"})
            new_req = request.clone(rel_url=query_url)
            await view_msg(request=new_req, ws=ws)
        elif cmd_method == "replay_msg":
            message_id = params[1]
            request.match_info["message_id"] = message_id
            await replay_msg(request=request, ws=ws)
        elif cmd_method == "list_msgs":
            query_params = {
                "start": params[1],
                "count": params[2],
                "order_by": params[3],
                "start_dt": params[4],
                "end_dt": params[5],
                "text": params[6],
                "rtext": params[7],
            }
            query_params = {k: v for k, v in query_params.items() if v is not None}
            query_url = request.rel_url.with_query(query_params)
            new_req = request.clone(rel_url=query_url)
            await list_msgs(request=new_req, ws=ws)
        elif cmd_method == "start_channel":
            await start_channel(request=request, ws=ws)
        elif cmd_method == "stop_channel":
            await stop_channel(request=request, ws=ws)
        else:
            await ws.send_str(f"{cmd_method} is not a valid method")



import asyncio

import pytest
import pytest_asyncio.plugin  # noqa F401

from pypeman import msgstore
from pypeman import channels
from pypeman.channels import BaseChannel
from pypeman.remoteadmin import RemoteAdminClient
from pypeman.remoteadmin import RemoteAdminServer
from pypeman.test import TearDownProjectTestCase as TestCase
from pypeman.tests.common import generate_msg
from pypeman.tests.common import TstNode


class RemoteAdminTests(TestCase):

    @pytest.fixture(autouse=True)
    def initfixture(self, unused_tcp_port):
        self.tcp_port = unused_tcp_port

    def clean_loop(self):
        # Useful to execute future callbacks
        pending = asyncio.all_tasks(loop=self.loop)

        if pending:
            self.loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))

    def start_channels(self):
        # Start channels
        for chan in channels.all_channels:
            self.loop.run_until_complete(chan.start())

    def setUp(self):
        # Create class event loop used for tests to avoid failing
        # previous tests to impact next test ? (Not sure)
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        # Remove thread event loop to be sure we are not using
        # another event loop somewhere
        asyncio.set_event_loop(None)

        # Avoid calling already tested channels
        channels.all_channels.clear()

    def tearDown(self):
        super().tearDown()
        self.clean_loop()

    def test_remote_admin_list(self):
        """ Channel remote listing working """

        port = self.tcp_port  # port used for rmt admin

        store_factory = msgstore.MemoryMessageStoreFactory()

        chan = BaseChannel(name="test_remote050", loop=self.loop, message_store_factory=store_factory)

        n = TstNode()
        n2 = TstNode(name="sub")
        n3 = TstNode(name="sub1")
        n4 = TstNode(name="sub2")

        msg = generate_msg(with_context=True)
        msg2 = generate_msg(timestamp=(1982, 11, 27, 12, 35), message_content="message content2")
        msg3 = generate_msg(timestamp=(1982, 11, 28, 12, 35), message_content="message_content3")
        msg4 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content="message content4")

        idref_msg3 = msg3.uuid

        chan.add(n)

        sub = chan.fork(name="subchannel")
        sub.append(n2, n3, n4)

        # Launch channel processing
        self.start_channels()
        self.loop.run_until_complete(chan.handle(msg))
        self.loop.run_until_complete(chan.handle(msg2))
        self.loop.run_until_complete(chan.handle(msg3))
        self.loop.run_until_complete(chan.handle(msg4))

        server = RemoteAdminServer(loop=self.loop, port=port)

        self.loop.run_until_complete(server.start())

        client = RemoteAdminClient(loop=self.loop, url="ws://localhost:%d" % port)
        client.init()

        # List channels
        chans = client.channels()

        print(chans)

        self.assertEqual(chans[0]['name'], 'test_remote050', "Channel listing not working")

        self.assertEqual(
            chans[0]['subchannels'][0]['name'],
            'test_remote050.subchannel',
            "Subchannel listing not working")

        # Stop channel
        result = client.stop('test_remote050')

        self.assertEqual(chan.status, BaseChannel.STOPPED, "Stopping channel doesn't work")

        # Start channel
        result = client.start('test_remote050')

        self.assertEqual(chan.status, BaseChannel.WAITING, "Starting channel doesn't work")

        # Search message
        msg_list = client.list_msgs(channel='test_remote050', start=2, count=5, order_by='-timestamp')

        print(msg_list)

        self.assertEqual(msg_list['total'], 4, 'List channel messages broken')
        self.assertEqual(msg_list['messages'][0]['id'], idref_msg3, 'List channel messages broken')

        # Search message with date filter
        msg_list = client.list_msgs(
            channel='test_remote050', start_dt="1982-11-27", end_dt="1982-11-28T13:00:00")

        print(msg_list)

        self.assertEqual(len(msg_list['messages']), 2, 'List channel messages broken')
        self.assertEqual(msg_list['messages'][1]['id'], idref_msg3, 'List channel messages broken')

        # Search message with text filter
        msg_list = client.list_msgs(
            channel='test_remote050', text="sage_c")

        print(msg_list)

        self.assertEqual(len(msg_list['messages']), 1, 'List channel messages broken')
        self.assertEqual(msg_list['messages'][0]['id'], idref_msg3, 'List channel messages broken')

        # Search message with regex filter
        msg_list = client.list_msgs(
            channel='test_remote050', rtext="\w+_\w+")  # noqa: W605

        print(msg_list)

        self.assertEqual(len(msg_list['messages']), 1, 'List channel messages broken')
        self.assertEqual(msg_list['messages'][0]['id'], idref_msg3, 'List channel messages broken')

        # Replay message
        result = client.replay_msg('test_remote050', idref_msg3)

        msg_list = client.list_msgs(channel='test_remote050', start=0, count=5, order_by='-timestamp')
        self.assertEqual(msg_list['total'], 5, 'List channel messages broken')
        self.assertEqual(msg_list['messages'][0]['id'], result.uuid, 'Replay messages broken')

        # Push message
        result = client.push_msg(channel='test_remote050', text="Yaaay")

        msg_list = client.list_msgs(channel='test_remote050', start=0, count=5, order_by='-timestamp')
        self.assertEqual(msg_list['total'], 6, 'Push message broken')
        self.assertEqual(msg_list['messages'][0]['id'], result.uuid, 'Push message broken')

        # View message
        msg_infos = client.view_msg(channel='test_remote050', msg_id=idref_msg3)
        print(msg_infos)

        self.assertEqual(msg_infos.payload, msg3.payload, 'View messages broken')

        # Preview message
        msg_infos = client.preview_msg(channel='test_remote050', msg_id=idref_msg3)
        print(msg_infos)

        self.assertEqual(msg_infos.payload, msg3.payload[:1000], 'Preview messages broken')



# pypeman -- A command-line argument parser for Python
# Copyright (C) 2015-2016 by MHComm. See LICENSE for details

import os
import sys
import unittest
import subprocess
import logging

from pypeman.helpers.tempfile import mktempfname

logger = logging.getLogger(__name__)

CWD = os.path.join('pypeman', 'tests', 'test_app')


class BinPypemanTestCase(unittest.TestCase):
    """ tests for the bin/pypeman cli """

    def setUp(self):
        """ prep a test """
        pypeman = os.path.join(os.path.dirname(__file__), '..', '..', 'pypeman', 'commands.py')
        self.cmd = [sys.executable, pypeman]
        os.environ['PYTHONPATH'] = os.path.join(os.path.dirname(__file__), '..', '..')
        self.tempfiles = []

    def tearDown(self):
        """ cleanup potentially created temp files """
        for fname in self.tempfiles:
            if os.path.exists(fname):
                os.unlink(fname)

    def run_pypeman(self, cmd, cwd=None):
        """
        Runs a command, gathers output and captures exit code
        :return: return_code and data (bytestring of stdout / stderr)
        """
        out_fname = mktempfname()
        self.tempfiles.append(out_fname)

        if cwd is None:
            cwd = os.getcwd()

        # TODO: why not reading stdout / stderr directoy into a var ??
        # TODO: if not reason can be found change code to use no temp file
        # TODO: perhaps some issue with unicode / python 3 ???
        with open(out_fname, 'wb') as fout:
            proc = subprocess.Popen(cmd, stdout=fout, stderr=fout, cwd=cwd)
            proc.wait()
        ret_code = proc.returncode

        data = None
        if ret_code:
            with open(out_fname, 'rb') as fin:
                data = fin.read()

        self.assertEqual(ret_code, 0, "exit code %d when calling %r in %s: %s" %
                         (ret_code, cmd, cwd, data))

        return ret_code, data

    def test_01_can_call_pypeman(self):
        """ pypeman can be called without params """
        logger.info("FILE = %r / NAME = %r", __file__, __name__)
        cmd = self.cmd
        self.run_pypeman(cmd)
        self.run_pypeman(cmd, cwd=CWD)

    def test_02_can_call_help(self):
        """ option -h is working """
        logger.info("FILE = %r / NAME = %r", __file__, __name__)

        cmd = self.cmd + ['-h']
        self.run_pypeman(cmd)
        self.run_pypeman(cmd, cwd=CWD)

    def test_03_can_call_graph(self):
        """ subcommand graph is working """

        cmd = self.cmd + ['graph']
        self.run_pypeman(cmd, cwd=CWD)

    def test_04_can_call_test(self):
        """ subcommand test is working """

        cmd = self.cmd + ['test']
        self.run_pypeman(cmd, cwd=os.path.join(os.path.dirname(CWD), 'test_app_testing'))

# test_suite =  BinPypemanTestCase



from pypeman.test import PypeTestCase
from pypeman import message


class Tests(PypeTestCase):

    def test_simple(self):
        msg = message.Message()
        msg.payload = "test"

        chan = self.get_channel('base')

        end = chan.handle_and_wait(msg)

        self.assertEqual(end.payload, 'test')



import importlib
import sys
import traceback


def load_class(module, class_, deps):
    """
    Try to load a class if all deps can be imported.
    :param module: path to module contains class
    :param class_:
    :param deps:
    :return:
    """
    try:
        mod = importlib.import_module(module)
        return getattr(mod, class_)
    except ImportError as exc:
        traceback.print_exc()
        msg = str(exc)

        # Try to find any dependency in message
        found = False
        for dep in deps:
            if dep in msg:
                found = True
                break

        if not found:
            print("IMPORT ERROR...", file=sys.stderr)
            raise

        print("%s module not activated" % module)
        raise


def load(selfmodname, module, class_, dep=None):
    """
    load a class and add it to selfmodname namespace.
    :param selfmodname:
    :param module:
    :param class_:
    :param dep:
    :return:
    """
    if dep is None:
        dep = []

    selfmod = importlib.import_module(selfmodname)

    def init(*args, **kwargs):
        C = load_class(module, class_, dep)
        setattr(selfmod, class_, C)
        return C(*args, **kwargs)

    return init


class Wrapper(object):
    def __init__(self, wrapped):

        self._wrapped = sys.modules[wrapped]
        self._extra = {}

        self.__dict__.update(self._wrapped.__dict__)

        sys.modules[wrapped] = self

    def add_lazy(self, module, name, deps):
        self._extra[name] = (module, deps)

    def __getattr__(self, name):
        try:
            return getattr(self._wrapped, name)
        except AttributeError:
            if name in self._extra:
                setattr(self._wrapped, name, load_class(self._extra[name][0], name, self._extra[name][1]))
                return getattr(self._wrapped, name)
            else:
                raise



import asyncio
import logging
import socket

logger = logging.getLogger(__name__)


all_endpoints = []


def reset_pypeman_endpoints():
    """
    clears book keeping of all endpoints

    Can be useful for unit testing.
    """
    all_endpoints.clear()


class BaseEndpoint:

    def __init__(self):
        all_endpoints.append(self)

    async def start(self):
        pass

    async def stop(self):
        pass


class SocketEndpoint(BaseEndpoint):
    def __init__(self, loop=None, sock=None, default_port='8080', reuse_port=None):
        """
            :param reuse_port: bool if true then the listening port specified in the url parameter)
                will be shared with other processes on same port
                no effect with bound socket object
            :param sock: string 'host:port'
                or socket-string ("unix:/sojet/file/path")
                or bound socket object
        """
        super().__init__()
        self.loop = loop or asyncio.get_event_loop()
        self.reuse_port = reuse_port

        self.sock = self.normalize_socket(sock, default_port=default_port)
        self.sock_obj = None

    @staticmethod
    def normalize_socket(sock, default_port="8080"):
        if isinstance(sock, str):
            if not sock.startswith('unix:'):
                if ':' not in sock:
                    sock += ':'
                host, port = sock.split(":")
                host = host or '127.0.0.1'
                port = port or default_port
                sock = host + ':' + port
        return sock

    @staticmethod
    def mk_socket(sock, reuse_port):
        """
            make, bind and return socket if string object is passed
            if not return sock as is
        """
        if isinstance(sock, str):
            logger.debug("trying to create socket %s", sock)
            if not sock.startswith('unix:'):
                try:
                    host, port = sock.split(":")
                    port = int(port)
                    if not host:
                        raise
                    bind_param = (host, port)
                except Exception:
                    logger.error('error on sock params in socket endpoint')
                    raise
                sock_obj = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock_obj.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            else:
                bind_param = sock.split(":", 1)[1]
                sock_obj = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            if reuse_port:
                SO_REUSEPORT = 15
                sock_obj.setsockopt(socket.SOL_SOCKET, SO_REUSEPORT, 1)
            sock_obj.bind(bind_param)
            logger.debug("socket %s created and binded", sock)
        else:
            logger.debug("no socket to create, keeping socket %s", repr(sock))
            sock_obj = sock
        return sock_obj

    def make_socket(self):
        """
            make and bind socket if string object is passed
        """
        self.sock_obj = self.mk_socket(self.sock, self.reuse_port)

    async def stop(self):
        if self.sock_obj:
            self.sock_obj.close()
            self.sock_obj = None


from pypeman.helpers import lazyload  # noqa: E402

wrap = lazyload.Wrapper(__name__)

wrap.add_lazy('pypeman.contrib.http', 'HTTPEndpoint', ['aiohttp'])

wrap.add_lazy('pypeman.contrib.hl7', 'MLLPEndpoint', ['hl7'])



# pypeman -- A command-line argument parser for Python
# Copyright (C) 2015-2016 by MHComm. See LICENSE for details

import logging
from logging.config import dictConfig

from pypeman.test import TearDownProjectTestCase as TestCase


class DefaultSettingsTests(TestCase):
    def test_01_default_settings(self):
        """ default settings exists """
        import pypeman.default_settings as dflt_settings  # noqa: F401

    def test_02_default_logging(self):
        """ default_logging is valid """
        import pypeman.default_settings as dflt_settings
        dictConfig(dflt_settings.LOGGING)
        logger = logging.getLogger(__name__)
        logger.debug("debug")
        logger.info("info")
        logger.warning("warning")
        logger.error("error")
        logger.critical("critical")



"""
Base class for a pypeman plugin

All interfaces are specified nothing is implemented
"""

import logging

from inspect import iscoroutinefunction


logger = logging.getLogger(__name__)


class BasePlugin:
    (INITIALIZED, READY, STARTING, STARTED, STOPPING, STOPPED, DESTROYED) = \
        states = range(7)

    STATE_NAMES = (
        "INITIALIZED", "READY", "STARTING", "STARTED",
        "STOPPING", "STOPPED", "DESTROYED",
        )

    def __init__(self):
        """
        Early initialisation before any pypeman endpoints and channels are created
        """
        self.name = f"{__name__}.{self.__class__}"
        logger.debug("Initialing plugin %s", self.name)
        self.status = self.INITIALIZED
        self.loop = None
        self.to_start = []
        self.start_results = []

    def do_ready(self):
        """
        Code to be executed after the whole pypeman graph has been generated.
        ready() functions of all plugins are executed before the loop is generated
        """
        logger.debug("called ready for plugin %s", self.name)
        assert self.status == self.INITIALIZED
        self.ready()
        self.status = self.READY

    def ready(self):
        pass

    def set_loop(self, loop):
        self.loop = loop

    def set_starting(self):
        logger.debug("starting plugin %s", self.name)
        assert self.status == self.READY
        self.status = self.STARTING

    def set_started(self):
        self.status = self.STARTED

    async def do_start(self):
        """
        Code to be executed after the creation of the event loop
        """
        self.set_starting()
        if iscoroutinefunction(self.start):
            logger.debug("awaiting coro start()")
            future = await self.start()
            logger.debug("started => %s", repr(future))
            self.start_results.append(future)
        else:
            start_rslt = self.start()
            logger.debug("start rslt = %s", repr(start_rslt))
            if isinstance(start_rslt, (list, tuple)):
                logger.debug("start_rslt is no coro")
                for coro in start_rslt:
                    rslt = await coro
                    self.start_results.append(rslt)
            elif start_rslt is None:
                logger.debug("start_rslt is None")
            elif iscoroutinefunction(start_rslt):
                logger.debug("start_rslt is a coro")
                rslt = await start_rslt()
                logger.debug("rslt: %s", repr(rslt))
                self.start_results.append(rslt)
            else:
                logger.debug("start_rslt is awaitable")
                rslt = await start_rslt
                logger.debug("rslt: %s", repr(rslt))
                self.start_results.append(rslt)
        self.set_started()
        logger.debug("end of start")

    async def start(self):
        pass

    def start_coro_list(self):
        """
        alternative start implementation returning a list of coros to be executed.
        """
        return [self.do_start()]

    async def do_stop(self):
        """
        Code to be executed before mainloop shall be stopped
        """
        logger.debug("stopping plugin %s", self.name)
        assert self.status in (self.INITIALIZED, self.STARTING, self.STARTED)
        if self.status == self.STARTING:
            # TODO: should change lateron, that a starting plugin can also
            #       be stopped
            raise RuntimeError("had to wait till started before stopping")
        elif self.status == self.STARTED:
            if iscoroutinefunction(self.stop):
                logger.debug("awaiting coro stop()")
                await self.stop()
            else:
                coro = self.stop()
                assert iscoroutinefunction(coro)
                await coro
            self.status = self.STOPPED

    async def stop(self):
        pass

    def do_destroy(self):
        """
        Code to be executed after all plugins are stopped
        """
        logger.debug("destroying plugin %s", self.name)
        assert self.status == self.STOPPED
        self.destroy()

    def destroy(self):
        assert self.status in (self.INITIALIZED, self.STOPPED)
        self.status = self.DESTROYED




import asyncio
import json
import os
import shutil
import tempfile

from pathlib import Path

from pypeman import channels
from pypeman import msgstore
from pypeman import nodes
from pypeman.channels import BaseChannel
from pypeman.test import TearDownProjectTestCase as TestCase
from pypeman.tests.common import generate_msg
from pypeman.tests.common import TstException
from pypeman.tests.common import TstNode


class ObjectWithoutStr:
    def __str__(self):
        raise NotImplementedError("__str__ not implemented")


class TstConditionalErrorNode(nodes.BaseNode):

    def process(self, msg):
        print("Process %s" % self.name)

        if isinstance(msg.payload, str) and "shall_fail" in msg.payload:
            raise TstException()

        return msg


class MsgstoreTests(TestCase):
    def clean_loop(self):
        # Useful to execute future callbacks
        pending = asyncio.all_tasks(loop=self.loop)

        if pending:
            self.loop.run_until_complete(asyncio.gather(*pending))

    def start_channels(self):
        # Start channels
        for chan in channels.all_channels:
            self.loop.run_until_complete(chan.start())

    def setUp(self):
        # Create class event loop used for tests to avoid failing
        # previous tests to impact next test ? (Not sure)
        self.loop = asyncio.new_event_loop()
        self.loop.set_debug(True)
        # Remove thread event loop to be sure we are not using
        # another event loop somewhere
        asyncio.set_event_loop(None)

        # Avoid calling already tested channels
        channels.all_channels.clear()

    def tearDown(self):
        super().tearDown()
        self.clean_loop()

    def test_null_message_store(self):
        """ We can store a message in NullMessageStore """

        chan = BaseChannel(
            name="test_channel9", loop=self.loop,
            message_store_factory=msgstore.FakeMessageStoreFactory())
        n = TstNode()
        msg = generate_msg(with_context=True)

        chan.add(n)

        # Launch channel processing
        self.start_channels()
        self.loop.run_until_complete(chan.handle(msg))

        self.assertTrue(n.processed, "Channel handle not working")

    def test_memory_message_store(self):
        """ We can store a message in FileMessageStore """

        store_factory = msgstore.MemoryMessageStoreFactory()

        chan = BaseChannel(name="test_channel10", loop=self.loop, message_store_factory=store_factory)

        n = TstNode()
        n_error = TstConditionalErrorNode()

        msg = generate_msg(with_context=True)
        msg2 = generate_msg(timestamp=(1982, 11, 27, 12, 35), message_content="message content2")
        msg3 = generate_msg(timestamp=(1982, 11, 28, 12, 35), message_content="message content3")
        msg4 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content="message_content4")

        # This message should be in error
        msg5 = generate_msg(timestamp=(1982, 11, 12, 14, 35),
                            message_content='{"test1": "shall_fail"}')
        msg6 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content=ObjectWithoutStr())

        chan.add(n)
        chan.add(n_error)

        # Launch channel processing
        self.start_channels()
        self.loop.run_until_complete(chan.handle(msg))
        self.loop.run_until_complete(chan.handle(msg2))
        self.loop.run_until_complete(chan.handle(msg3))
        self.loop.run_until_complete(chan.handle(msg4))

        with self.assertRaises(TstException):
            # This message should be in error state
            self.loop.run_until_complete(chan.handle(msg5))
        self.loop.run_until_complete(chan.handle(msg6))
        msg6 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content=ObjectWithoutStr())

        self.assertEqual(self.loop.run_until_complete(chan.message_store.total()),
                         6, "Should be a total of 6 messages in store!")

        msg_stored1 = list(self.loop.run_until_complete(chan.message_store.search(0, 2)))
        self.assertEqual(len(msg_stored1), 2, "Should be 2 results from search!")

        msg_stored2 = list(self.loop.run_until_complete(chan.message_store.search(2, 5)))
        self.assertEqual(len(msg_stored2), 4, "Should be 4 results from search!")

        msg_stored = list(self.loop.run_until_complete(chan.message_store.search()))

        # All message stored ?
        self.assertEqual(len(msg_stored), 6, "Should be 6 messages in store!")

        for msg in msg_stored:
            print(msg)

        for msg in msg_stored1 + msg_stored2:
            print(msg)

        ids1 = [msg['id'] for msg in msg_stored1 + msg_stored2]
        ids2 = [msg['id'] for msg in msg_stored]

        self.assertEqual(ids1, ids2, "Should be 6 messages in store!")

        # Test processed message
        dict_msg = self.loop.run_until_complete(chan.message_store.get('%s' % msg3.uuid))
        self.assertEqual(dict_msg['state'], 'processed', "Message %s should be in processed state!" % msg3)

        # Test failed message
        dict_msg = self.loop.run_until_complete(chan.message_store.get('%s' % msg5.uuid))
        self.assertEqual(dict_msg['state'], 'error', "Message %s should be in error state!" % msg5)

        # Test list messages
        msgs = self.loop.run_until_complete(chan.message_store.search(start=2, count=5))
        self.assertEqual(len(msgs), 4, "Failure of listing messages from memory msg store")

        # Test list messages with date filters
        msgs = self.loop.run_until_complete(chan.message_store.search(
            start_dt="1982-11-27", end_dt="1982-11-28T13:00:00"))
        self.assertEqual(len(msgs), 2, "Failure of listing messages from memory msg store")

        # Test list messages with text filter
        msgs = self.loop.run_until_complete(chan.message_store.search(text="sage con"))
        self.assertEqual(len(msgs), 2, "Failure of listing messages from memory msg store")

        # Test view message
        msg_content = self.loop.run_until_complete(chan.message_store.get_msg_content('%s' % msg5.uuid))
        self.assertEqual(msg_content.payload, msg5.payload, "Failure of message %s view!" % msg5)

        # Test preview message
        msg_content = self.loop.run_until_complete(chan.message_store.get_preview_str('%s' % msg5.uuid))
        self.assertEqual(msg_content.payload, msg5.payload[:1000], "Failure of message %s preview!" % msg5)

    def test_memory_message_store_in_fork(self):
        """ We can store a message in FileMessageStore """

        store_factory = msgstore.MemoryMessageStoreFactory()

        chan = BaseChannel(name="test_channel10.25", loop=self.loop, message_store_factory=store_factory)

        n1 = TstNode()
        n2 = TstNode()
        n3 = TstNode()
        n4 = TstNode()

        chan.add(n1, n2)
        fork = chan.fork()
        fork.add(n3)

        self.assertTrue(isinstance(fork.message_store, msgstore.NullMessageStore))

        whe = chan.when(True, message_store_factory=store_factory)
        whe.add(n4)

        self.assertTrue(isinstance(whe.message_store, msgstore.MemoryMessageStore))

    def test_replay_from_memory_message_store(self):
        """ We can store a message in FileMessageStore """

        store_factory = msgstore.MemoryMessageStoreFactory()

        chan = BaseChannel(name="test_channel10.5", loop=self.loop, message_store_factory=store_factory)

        n = TstNode()

        msg = generate_msg(with_context=True)
        msg2 = generate_msg(timestamp=(1982, 11, 27, 12, 35))

        chan.add(n)

        # Launch channel processing
        self.start_channels()
        self.loop.run_until_complete(chan.handle(msg))
        self.loop.run_until_complete(chan.handle(msg2))

        msg_stored = list(self.loop.run_until_complete(chan.message_store.search()))

        for msg in msg_stored:
            print(msg)

        self.assertEqual(len(msg_stored), 2, "Should be 2 messages in store!")

        self.loop.run_until_complete(chan.replay(msg_stored[0]['id']))

        msg_stored = list(self.loop.run_until_complete(chan.message_store.search()))
        self.assertEqual(len(msg_stored), 3, "Should be 3 messages in store!")

    def test_file_message_store(self):
        """ We can store a message in FileMessageStore """

        tempdir = tempfile.mkdtemp()
        print(tempdir)

        store_factory = msgstore.FileMessageStoreFactory(path=tempdir)

        chan = BaseChannel(name="test_channel11", loop=self.loop, message_store_factory=store_factory)

        n = TstNode()
        n_error = TstConditionalErrorNode()

        msg = generate_msg(with_context=True)
        msg2 = generate_msg(timestamp=(1982, 11, 27, 12, 35), message_content="message content2")
        msg3 = generate_msg(timestamp=(1982, 11, 28, 12, 35), message_content="message content3")
        msg4 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content="message_content4")

        # This message should be in error
        msg5 = generate_msg(timestamp=(1982, 11, 12, 14, 35),
                            message_content='{"test1": "shall_fail"}')
        msg6 = generate_msg(timestamp=(1982, 11, 28, 14, 35), message_content=ObjectWithoutStr())

        chan.add(n)
        chan.add(n_error)

        # Launch channel processing
        self.start_channels()
        self.loop.run_until_complete(chan.handle(msg))
        self.loop.run_until_complete(chan.handle(msg2))
        self.loop.run_until_complete(chan.handle(msg3))
        self.loop.run_until_complete(chan.handle(msg4))

        with self.assertRaises(TstException):
            # This message should be in error state
            self.loop.run_until_complete(chan.handle(msg5))
        self.loop.run_until_complete(chan.handle(msg6))

        self.assertEqual(self.loop.run_until_complete(chan.message_store.total()),
                         6, "Should be a total of 6 messages in store!")

        msg_stored1 = list(self.loop.run_until_complete(chan.message_store.search(0, 2)))
        self.assertEqual(len(msg_stored1), 2, "Should be 2 results from search!")

        msg_stored2 = list(self.loop.run_until_complete(chan.message_store.search(2, 5)))
        self.assertEqual(len(msg_stored2), 4, "Should be 4 results from search!")

        msg_stored = list(self.loop.run_until_complete(chan.message_store.search()))

        # All message stored ?
        self.assertEqual(len(msg_stored), 6, "Should be 6 messages in store!")

        for msg in msg_stored:
            print(msg)

        for msg in msg_stored1 + msg_stored2:
            print(msg)

        ids1 = [msg['id'] for msg in msg_stored1 + msg_stored2]
        ids2 = [msg['id'] for msg in msg_stored]

        self.assertEqual(ids1, ids2, "Should be 5 messages in store!")

        # Test processed message
        dict_msg = self.loop.run_until_complete(
            chan.message_store.get('19821128_1235_%s' % msg3.uuid))
        self.assertEqual(dict_msg['state'], 'processed', "Message %s should be in processed state!" % msg3)

        # Test failed message
        dict_msg = self.loop.run_until_complete(
            chan.message_store.get('19821112_1435_%s' % msg5.uuid))
        self.assertEqual(dict_msg['state'], 'error', "Message %s should be in error state!" % msg5)

        self.assertTrue(os.path.exists("%s/%s/1982/11/28/19821128_1235_%s"
                        % (tempdir, chan.name, msg3.uuid)))

        # Test list messages
        msgs = self.loop.run_until_complete(chan.message_store.search(start=2, count=5))
        self.assertEqual(len(msgs), 4, "Failure of listing messages for file msg store")

        # Test list messages with date filters
        msgs = self.loop.run_until_complete(chan.message_store.search(
            start_dt="1982-11-27", end_dt="1982-11-28T13:00:00"))
        self.assertEqual(len(msgs), 2, "Failure of listing messages for file msg store")

        # Test list messages with text filters
        msgs = self.loop.run_until_complete(chan.message_store.search(
            text="sage con"))
        self.assertEqual(len(msgs), 2, "Failure of listing messages for file msg store")

        # Test list messages with regex filters
        msgs = self.loop.run_until_complete(chan.message_store.search(
            rtext="\w+_\w+"))  # noqa: W605
        self.assertEqual(len(msgs), 1, "Failure of listing messages for file msg store")

        # Test view message
        msg_content = self.loop.run_until_complete(chan.message_store.get_msg_content(
            '19821112_1435_%s' % msg5.uuid))
        self.assertEqual(msg_content.payload, msg5.payload, "Failure of message %s view!" % msg5)

        # Test preview message
        msg_content = self.loop.run_until_complete(chan.message_store.get_preview_str(
            '19821112_1435_%s' % msg5.uuid))
        self.assertEqual(msg_content.payload, msg5.payload[:1000], "Failure of message %s preview!" % msg5)

        self.clean_loop()

        # TODO put in tear_down ?
        shutil.rmtree(tempdir, ignore_errors=True)

    def test_file_message_store_meta(self):
        """Tests for meta in file message store"""
        data_tst_path = Path(__file__).resolve().parent / "data"
        old_meta_tst_path = data_tst_path / "old_meta.meta"
        new_meta_tst_path = data_tst_path / "new_meta.meta"
        with new_meta_tst_path.open("r") as fin:
            new_meta_data = json.load(fin)
        msg_uid = "msgid"
        msg_year = "2024"
        msg_month = "06"
        msg_day = "13"
        msg_date = f"{msg_year}{msg_month}{msg_day}"
        msg_time = "0000"
        msg_id = f"{msg_date}_{msg_time}_{msg_uid}"

        with tempfile.TemporaryDirectory() as tempdir:
            store_factory = msgstore.FileMessageStoreFactory(path=tempdir)
            store = store_factory.get_store(store_id="")
            meta_dst_folder_path = Path(tempdir) / msg_year / msg_month / msg_day
            meta_dst_folder_path.mkdir(parents=True, exist_ok=True)
            meta_dst_path = meta_dst_folder_path / f"{msg_id}.meta"
            shutil.copy(old_meta_tst_path, meta_dst_path)
            # Tests that msgstore could read an old meta file
            msg_state = asyncio.run(store.get_message_state(msg_id))
            self.assertEqual(msg_state, "processed")
            # Test that the old meta file is converted to a new json meta file
            with meta_dst_path.open("r") as fin:
                meta_data = json.load(fin)
            self.assertDictEqual(new_meta_data, meta_data)



import logging

from unittest import mock

from .common import generate_msg

from pypeman.message import Message
from pypeman.test import TearDownProjectTestCase as TestCase


class MessageTests(TestCase):

    def test_message_dict_conversion(self):
        m = generate_msg(message_content={'answer': 42}, with_context=True)

        mdict = m.to_dict()

        self.assertTrue(isinstance(mdict, dict), "Message as_dict method is broken")

        compare_to = Message.from_dict(mdict)

        self.assertEqual(m.payload['answer'], compare_to.payload['answer'], "Payload not well decoded")
        self.assertEqual(m.uuid, compare_to.uuid, "Bad uuid")
        self.assertEqual(m.meta['question'], compare_to.meta['question'], "Bad meta")

        self.assertEqual(
            m.ctx['test']['payload']['question'],
            compare_to.ctx['test']['payload']['question'], "Bad ctx")
        self.assertEqual(
            m.ctx['test']['meta']['answer'],
            compare_to.ctx['test']['meta']['answer'], "Bad ctx")

    def test_message_json_conversion(self):
        m = generate_msg(message_content={'answer': 42}, with_context=True)

        msg_json = m.to_json()

        compare_to = Message.from_json(msg_json)

        self.assertEqual(m.payload['answer'], compare_to.payload['answer'], "Payload not well decoded")
        self.assertEqual(m.uuid, compare_to.uuid, "Bad uuid")
        self.assertEqual(m.meta['question'], compare_to.meta['question'], "Bad meta")

        self.assertEqual(
            m.ctx['test']['payload']['question'],
            compare_to.ctx['test']['payload']['question'], "Bad ctx")
        self.assertEqual(
            m.ctx['test']['meta']['answer'],
            compare_to.ctx['test']['meta']['answer'], "Bad ctx")

    def test_message_copy(self):
        m = generate_msg(message_content={'answer': 42}, with_context=True)

        compare_to = m.copy()

        self.assertEqual(m.payload['answer'], compare_to.payload['answer'], "Payload not well copied")
        self.assertEqual(m.uuid, compare_to.uuid, "Bad uuid copy")
        self.assertEqual(m.meta['question'], compare_to.meta['question'], "Bad meta copy")

        self.assertEqual(
            m.ctx['test']['payload']['question'],
            compare_to.ctx['test']['payload']['question'], "Bad ctx copy")
        self.assertEqual(
            m.ctx['test']['meta']['answer'],
            compare_to.ctx['test']['meta']['answer'], "Bad ctx")

    def test_message_renew(self):
        m = generate_msg(message_content={'answer': 42}, with_context=True)

        compare_to = m.renew()

        self.assertEqual(m.payload['answer'], compare_to.payload['answer'], "Payload not well copied")
        self.assertNotEqual(m.uuid, compare_to.uuid, "Uuid should not be copied")
        self.assertNotEqual(m.timestamp, compare_to.timestamp, "Timestamp should not be copied")
        self.assertEqual(m.meta['question'],
                         compare_to.meta['question'], "Bad meta copy")

        self.assertEqual(
            m.ctx['test']['payload']['question'],
            compare_to.ctx['test']['payload']['question'], "Bad ctx copy")
        self.assertEqual(m.ctx['test']['meta']['answer'],
                         compare_to.ctx['test']['meta']['answer'], "Bad ctx")

    def test_message_logging(self):
        """
        Whether message logging is working well.
        """
        m = generate_msg(message_content={'answer': 42}, with_context=True)

        mock_logger = mock.MagicMock()
        m.log(logger=mock_logger)

        mock_logger.log.assert_called_with(logging.DEBUG, 'Meta: %r', {'question': 'unknown'})
        mock_logger.reset()

        m.log(logger=mock_logger, payload=True, meta=True, context=True, log_level=logging.INFO)
        mock_logger.log.assert_called_with(logging.INFO, 'Meta: %r', {'answer': 43})

    def test_message_print(self):
        """
        Whether message print version is working well.
        """
        m = generate_msg(
            message_content={'answer': 42},
            with_context=True,
            timestamp=(2010, 10, 10, 10, 10, 10),
        )
        m.uuid = "msguuid"

        result = m.to_print(context=True)

        reference = """Message msguuid
Date: 2010-10-10 10:10:10
Payload: {'answer': 42}
Meta: {'question': 'unknown'}
Context for message ->
-- Key "test" --
Payload: {'question': 'known'}
Meta: {'answer': 43}
"""

        self.assertEqual(result, reference, "Print message not working")



import os
import tempfile


def mktempfname(suffix='', prefix='tmp', dir=None):
    """ creates a temporary file name
        in order to be nice to windows the file handle is closed immediately
    """
    fdesc, fname = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir)
    os.close(fdesc)
    return fname



"""
The Plugin Manager shall handle pypeman plugins.

The role of pypeman plugins is to allow inserting plugins, which don't really
make part of the pypeman graph, but are intended for debugging, monitoring
or any other code, that shall be present along the pypeman code.

It can also encourage development of plugins outside of pypeman, that can lateron
be merged into to pypeman.contrib.plugins or pypeman.plugins.
"""

import logging

from importlib import import_module

from pypeman.conf import settings


logger = logging.getLogger(__name__)


class PluginManager():
    def __init__(self):
        self.imported = False
        self.plugin_classes = []  # list of plugin modules
        self.plugins = []
        self.loop = None

    def set_loop(self, loop):
        self.loop = loop
        for plugin in self.plugins:
            plugin.set_loop(loop)

    def import_plugins(self):
        """
        import plugin modules and store the classes in a list
        """
        if self.imported:
            return
        for plugin_name in settings.PLUGINS:
            module_name, cls = plugin_name.rsplit(".", 1)
            module = import_module(module_name)
            plugin = getattr(module, cls)
            self.plugin_classes.append(plugin)

    def init_plugins(self):
        """
        initialize all plugins
        """
        # shutdown previous plugins if existing
        if self.plugins:
            for plugin in self.plugins:
                if plugin.status == plugin.STARTED:
                    plugin.do_stop()
            for plugin in self.plugins:
                if plugin.status == plugin.STOPPED:
                    plugin.do_destroy()

        # instantiate plugins
        self.plugins = []
        for plugin_cls in self.plugin_classes:
            self.plugins.append(plugin_cls())

    def ready_plugins(self):
        """
        call ready functions for all plugins.
        We know now, that *all* plugins are already
        instantiated
        """
        for plugin, plugin_cls in zip(self.plugins, self.plugin_classes):
            logger.info("calling ready for plugin %s", repr(plugin_cls))
            plugin.do_ready()

    def start_plugins(self):
        """
        call the start function for each plugin
        """
        loop = self.loop
        for plugin, plugin_cls in zip(self.plugins, self.plugin_classes):
            logger.info("starting plugin %s", repr(plugin_cls))
            loop.run_until_complete(plugin.do_start())
            # start whatever the plugin wants to start
            for to_start in plugin.to_start:
                loop.run_until_complete(to_start)

    def stop_plugins(self):
        """
        call the stop function for each plugin
        """
        loop = self.loop
        for plugin, plugin_cls in zip(self.plugins, self.plugin_classes):
            logger.info("stopping plugin %s", repr(plugin_cls))
            loop.run_until_complete(plugin.do_stop())


manager = PluginManager()



def flatten(items):
    """ Yield items from any nested iterable. """
    for x in items:
        if isinstance(x, (list, tuple)) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x



import datetime
import dateutil.parser
import json
import logging
import os
import re

from itertools import islice
from collections import OrderedDict
from pathlib import Path

from pypeman.message import Message

from pypeman.errors import PypemanConfigError

logger = logging.getLogger("pypeman.store")

DATE_FORMAT = '%Y%m%d_%H%M'


class MessageStoreFactory():
    """ Message store factory class can generate Message store instance for specific store_id. """

    def get_store(self, store_id):
        """
        :param store_id: identifier of corresponding message store.
        :return: A MessageStore corresponding to correct store_id.
        """


class MessageStore():
    """ A MessageStore keep an history of processed messages. Mainly used in channels. """

    async def start(self):
        """
        Called at startup to initialize store.
        """

    async def store(self, msg):
        """
        Store a message in the store.

        :param msg: The message to store.
        :return: Id for this specific message.
        """

    async def add_message_meta_infos(self, id, meta_info_name, info):
        """
        Add message meta infos in the store

        :param id: Message specific store id.
        :param meta_info_name: The name of the meta info to create/update
        :param info: The info value
        """

    async def get_message_meta_infos(self, id, meta_info_name=None):
        """
        Get message meta infos in the store

        :param id: Message specific store id.
        :param meta_info_name: (optional) The name of the meta info to get,
            if not set returns the entire dict
        """

    async def change_message_state(self, id, new_state):
        """
        Change the `id` message state.

        :param id: Message specific store id.
        :param new_state: Target state.
        """

    async def get(self, id):
        """
        Return one message corresponding to given `id` with his status.

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """

    async def get_preview_str(self, id):
        """
        Return the first 1000 chars of message content corresponding to `id`

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'message_content': str}`.
        """

    async def get_msg_content(self, id):
        """
        Return the content of the message corresponding to given `id`

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'message_content': Message}`.
        """

    async def is_regex_in_msg(self, id, rtext):
        """
        Return True if the str(msg) contains the regex rtext

        :param id: Message id. Message store dependant.
        :param rtext: string of regular expression to search in msg
        :return: True if it matches False otherwise
        """

    async def is_txt_in_msg(self, id, text):
        """
        Return True if the str(msg) contains param text

        :param id: Message id. Message store dependant.
        :param text: String. The text to search in msg
        :return: True if it text is found, False otherwise
        """

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):
        """
        Return a list of message with store specific `id` and processed status.

        :param start: First element.
        :param count: Count of elements since first element.
        :param order_by: Message order. Allowed values : ['timestamp', 'status'].
        :param start_dt: (optional) Isoformat start date(time) to filter with
        :param end_dt: (optional) Isoformat end date(time) to filter with
        :param text: (optional) String to search in message content
        :param rtext: (optional) String regex to search in message content
        :return: A list of dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """

    async def total(self):
        """
        :return: total count of messages
        """

    async def delete(self, id):
        """
        Delete one message in the store corresponding to given `id` with his status.
        Useful for tests and maybe in the future to permits cleanup of message store

        !CAUTION! : cannot be undone

        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """


class NullMessageStoreFactory(MessageStoreFactory):
    """ Return an NullMessageStore that do nothing at all. """
    def get_store(self, store_id):
        return NullMessageStore()


class NullMessageStore(MessageStore):
    """ For testing purpose """

    async def store(self, msg):
        return None

    async def get(self, id):
        return None

    async def add_message_meta_infos(self, id, meta_info_name, info):
        return None

    async def get_message_meta_infos(self, id, meta_info_name=None):
        return None

    async def get_preview_str(self, id):
        return None

    async def get_msg_content(self, id):
        return None

    async def search(self, **kwargs):
        return None

    async def total(self):
        return 0

    async def delete(self, id):
        return None


class FakeMessageStoreFactory(MessageStoreFactory):
    """ Return an Fake message store """
    def get_store(self, store_id):
        return FakeMessageStore()


class FakeMessageStore(MessageStore):
    """ For testing purpose """

    async def store(self, msg):
        logger.debug("Should store message %s", msg)
        return 'fake_id'

    async def add_message_meta_infos(self, id, meta_info_name, info):
        return None

    async def get_message_meta_infos(self, id, meta_info_name=None):
        return "processed"

    async def get(self, id):
        return {'id': id, 'state': 'processed', 'message': None}

    async def get_preview_str(self, id):
        return {"id": id, "message_content": "content"}

    async def get_msg_content(self, id):
        return {"id": id, "message_content": "content"}

    async def is_regex_in_msg(self, id, rtext):
        return True

    async def is_txt_in_msg(self, id, text):
        return True

    async def search(self, **kwargs):
        return []

    async def total(self):
        return 0

    async def delete(self, id):
        """
            we delete nothing here, but return what would have been deleted
            (for testing)
        """
        return {'id': id, 'state': 'processed', 'message': None}


class MemoryMessageStoreFactory(MessageStoreFactory):
    """ Return a Memory message store. All message are lost at pypeman stop. """
    def __init__(self):
        self.base_dict = {}

    def get_store(self, store_id):
        return MemoryMessageStore(self.base_dict, store_id)


class MemoryMessageStore(MessageStore):
    """ Store messages in memory """

    def __init__(self, base_dict, store_id):
        super().__init__()
        self.messages = base_dict.setdefault(store_id, OrderedDict())

    async def store(self, msg):
        msg_id = msg.uuid
        self.messages[msg_id] = {
            'id': msg_id, 'state': Message.PENDING,
            'timestamp': msg.timestamp, 'message': msg.to_dict()}
        return msg_id

    async def change_message_state(self, id, new_state):
        await self.add_message_meta_infos(id, "state", new_state)

    async def add_message_meta_infos(self, id, meta_info_name, info):
        self.messages[id][meta_info_name] = info

    async def get_message_meta_infos(self, id, meta_info_name=None):
        return self.messages[id][meta_info_name]

    async def get(self, id):
        resp = dict(self.messages[id])
        resp['message'] = Message.from_dict(resp['message'])
        return resp

    async def get_preview_str(self, id):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)[:1000]
        except Exception:
            msg.payload = repr(msg.payload)[:1000]
        return msg

    async def get_msg_content(self, id):
        msg = await self.get(id)
        msg_content = msg["message"]
        return msg_content

    async def is_regex_in_msg(self, id, rtext):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        regex = re.compile(rtext)
        return True if regex.match(msg.payload) else False

    async def is_txt_in_msg(self, id, text):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        return text in msg.payload

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):

        if order_by.startswith('-'):
            reverse = True
            sort_key = order_by[1:]
        else:
            reverse = False
            sort_key = order_by

        if start_dt:
            start_dt = dateutil.parser.isoparse(start_dt)
        if end_dt:
            end_dt = dateutil.parser.isoparse(end_dt)

        result = []
        values = (
            val for val in self.messages.values()
            if (not start_dt or val["timestamp"] >= start_dt)
            and (not end_dt or val["timestamp"] <= end_dt)
        )
        if text:
            found_values = []
            for val in values:
                if await self.is_txt_in_msg(val["id"], text):
                    found_values.append(val)
            values = found_values

        if rtext:
            found_values = []
            for val in values:
                if await self.is_regex_in_msg(val["id"], rtext):
                    found_values.append(val)
            values = found_values
        for value in islice(sorted(values, key=lambda x: x[sort_key], reverse=reverse),
                            start, start + count):
            resp = dict(value)
            resp['message'] = Message.from_dict(resp['message'])
            result.append(resp)
        return result

    async def total(self):
        return len(self.messages)

    async def delete(self, id):
        resp = dict(self.messages.pop(id))
        resp['message'] = Message.from_dict(resp['message'])
        return resp


class FileMessageStoreFactory(MessageStoreFactory):
    """
    Generate a FileMessageStore message store instance.
    Store a file in `<base_path>/<store_id>/<year>/<month>/<day>/` hierachy.
    """

    # TODO add an option to reguraly archive old file or delete them
    def __init__(self, path):
        super().__init__()
        if path is None:
            raise PypemanConfigError('file message store requires a path')
        self.base_path = path

    def get_store(self, store_id):
        return FileMessageStore(self.base_path, store_id)


class FileMessageStore(MessageStore):
    """ Store a file in `<base_path>/<store_id>/<year>/<month>/<day>/` hierachy."""
    # TODO file access should be done in another thread. Waiting for file backend.

    def __init__(self, path, store_id):
        super().__init__()

        self.base_path = os.path.join(path, store_id)

        # Match msg file name
        self.msg_re = re.compile(
            r'^(?P<msg_date>[0-9]{8})_(?P<msg_time>[0-9]{2}[0-9]{2})_(?P<msg_uid>[0-9a-zA-Z]*)$')

        try:
            # Try to make dirs if necessary
            os.makedirs(os.path.join(self.base_path))
        except FileExistsError:
            pass

        self._total = 0

    def id2path(self, id):
        match = self.msg_re.match(id)
        if not match:
            raise ValueError(f"Id '{id}' not a correct id")
        msg_str_date = match.groupdict()["msg_date"]
        year = msg_str_date[:4]
        month = msg_str_date[4:6]
        day = msg_str_date[6:8]
        msg_path = Path(self.base_path) / year / month / day / id
        return msg_path

    async def start(self):
        self._total = await self.count_msgs()

    async def store(self, msg):
        """ Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy."""
        # TODO implement a safer store to avoid broken messages

        # The filename is the file id
        filename = "{}_{}".format(msg.timestamp.strftime(DATE_FORMAT), msg.uuid)
        msg_path = self.id2path(filename)
        msg_path.parent.mkdir(parents=True, exist_ok=True)

        # Write message to file
        with msg_path.open("w") as f:
            f.write(msg.to_json())

        await self.change_message_state(filename, Message.PENDING)

        self._total += 1

        return filename

    def _is_json_meta(self, id):
        """Check if the message meta file is a json. If it's not,
        the message is an old message that doesnt contain other infos in
        meta except the state

        Args:
            id (str): The id of the message
        """
        msg_fpath = self.id2path(id)
        meta_fpath = msg_fpath.with_suffix(".meta")
        if not meta_fpath.exists():
            return False
        with meta_fpath.open("r") as fin:
            try:
                json.load(fin)
                return True
            except json.JSONDecodeError:
                return False

    def _convert_meta_to_json(self, id):
        """Convert an old message meta to a new one (json)

        Args:
            id (str): The id of the message to convert
        """
        msg_fpath = self.id2path(id)
        meta_fpath = msg_fpath.with_suffix(".meta")
        with meta_fpath.open("r") as fin:
            state = fin.read()
        meta_data = {"state": state}
        with meta_fpath.open("w") as fout:
            json.dump(meta_data, fout)
        return meta_data

    async def get_message_meta_infos(self, id, meta_info_name=None):
        msg_fpath = self.id2path(id)
        meta_fpath = msg_fpath.with_suffix(".meta")
        if meta_fpath.exists():
            is_new_meta_file = self._is_json_meta(id)
            if not is_new_meta_file:
                meta_data = self._convert_meta_to_json(id)
            else:
                with meta_fpath.open("r") as fin:
                    meta_data = json.load(fin)
        else:
            meta_data = {}

        if meta_info_name:
            meta_data = meta_data.get(meta_info_name)
        return meta_data

    async def add_message_meta_infos(self, id, meta_info_name, info):
        msg_fpath = self.id2path(id)
        meta_fpath = msg_fpath.with_suffix(".meta")
        if meta_fpath.exists():
            is_new_meta_file = self._is_json_meta(id)
            if not is_new_meta_file:
                meta_data = self._convert_meta_to_json(id)
            else:
                with meta_fpath.open("r") as fin:
                    meta_data = json.load(fin)
        else:
            meta_data = {}

        meta_data[meta_info_name] = info
        with meta_fpath.open("w") as fout:
            json.dump(meta_data, fout)

    async def change_message_state(self, id, new_state):
        await self.add_message_meta_infos(id, "state", new_state)

    async def get_message_state(self, id):
        return await self.get_message_meta_infos(id, "state")

    async def get(self, id):
        fpath = self.id2path(id)
        if not fpath.exists():
            raise IndexError

        with fpath.open("rb") as f:
            msg = Message.from_json(f.read().decode('utf-8'))
            return {
                'id': id,
                'state': await self.get_message_state(id),
                'message': msg,
                "meta": await self.get_message_meta_infos(id)
            }

    async def get_preview_str(self, id):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)[:1000]
        except Exception:
            msg.payload = repr(msg.payload)[:1000]
        return msg

    async def get_msg_content(self, id):
        msg = await self.get(id)
        msg_content = msg["message"]
        return msg_content

    async def sorted_list_directories(self, path, reverse=True):
        """
        :param path: Base path
        :param reverse: reverse order
        :return: List of directories in specified path ordered
        """
        return sorted([d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))], reverse=reverse)

    async def count_msgs(self):
        """
        Count message by listing all directories. To be used at startup.
        """
        count = 0
        for year in await self.sorted_list_directories(os.path.join(self.base_path)):
            for month in await self.sorted_list_directories(os.path.join(self.base_path, year)):
                for day in await self.sorted_list_directories(os.path.join(self.base_path, year, month)):
                    for msg_id in sorted(os.listdir(os.path.join(self.base_path, year, month, day))):
                        found = self.msg_re.match(msg_id)
                        if found:
                            count += 1
        return count

    async def is_regex_in_msg(self, id, rtext):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        regex = re.compile(rtext)
        return True if regex.match(msg.payload) else False

    async def is_txt_in_msg(self, id, text):
        msg = await self.get_msg_content(id)
        try:
            msg.payload = str(msg.payload)
        except Exception:
            msg.payload = repr(msg.payload)
        return text in msg.payload

    async def search(self, start=0, count=10, order_by='timestamp', start_dt=None, end_dt=None,
                     text=None, rtext=None):
        # TODO better performance for slicing by counting file in dirs ?
        if order_by.startswith('-'):
            reverse = True
            # sort_key = order_by[1:]
        else:
            reverse = False
            # sort_key = order_by

        if start_dt:
            start_dt = dateutil.parser.isoparse(start_dt)
        if end_dt:
            end_dt = dateutil.parser.isoparse(end_dt)

        # TODO handle sort_key
        result = []
        end = start + count
        position = 0
        for year in await self.sorted_list_directories(
                os.path.join(self.base_path), reverse=reverse):
            for month in await self.sorted_list_directories(
                    os.path.join(self.base_path, year), reverse=reverse):
                for day in await self.sorted_list_directories(
                        os.path.join(self.base_path, year, month), reverse=reverse):
                    # Pre filter with date to avoid listing unuseful dirs
                    msg_date = datetime.date(year=int(year), month=int(month), day=int(day))
                    if start_dt:
                        if msg_date < start_dt.date():
                            continue
                    if end_dt:
                        if msg_date > end_dt.date():
                            continue
                    for msg_id in sorted(
                            os.listdir(os.path.join(
                                self.base_path, year, month, day)),
                            reverse=reverse):
                        found = self.msg_re.match(msg_id)
                        if found:
                            msg_str_time = found.groupdict()["msg_time"]
                            hour = int(msg_str_time[:2])
                            minute = int(msg_str_time[2:4])
                            msg_time = datetime.time(hour=hour, minute=minute, second=0)
                            msg_dt = datetime.datetime.combine(msg_date, msg_time)
                            if start_dt:
                                if msg_dt < start_dt:
                                    continue
                            if end_dt:
                                if msg_dt > end_dt:
                                    continue
                            if text:
                                if not await self.is_txt_in_msg(msg_id, text):
                                    continue
                            if rtext:
                                if not await self.is_regex_in_msg(msg_id, rtext):
                                    continue
                            if start <= position < end:
                                # TODO: need to do processing of payload
                                #       before filtering (HL7 / json-str)
                                # TODO: add filter here
                                # TODO: can we transfoer into a generator?
                                result.append(await self.get(msg_id))
                            position += 1
        return result

    async def total(self):
        return self._total

    async def _delete_meta_file(self, id):
        msg_path = self.id2path(id)
        meta_fpath = msg_path.with_suffix(".meta")
        meta_fpath.unlink()

    async def delete(self, id):
        fpath = self.id2path(id)
        if not fpath.exists():
            raise IndexError

        with fpath.open("rb") as f:
            msg = Message.from_json(f.read().decode('utf-8'))

        data_to_return = {'id': id, 'state': await self.get_message_state(id), 'message': msg}
        fpath.unlink()
        self._delete_meta_file(id=id)
        return data_to_return
